import numbers
from enum import Enum

import numpy as np
from hydro_serving_grpc.serving.contract.types_pb2 import (
    DT_HALF, DT_FLOAT, DT_DOUBLE, DT_INT8, DT_INT16, DT_INT32, DT_INT64, DT_UINT8,
    DT_UINT16, DT_UINT32, DT_UINT64, DT_COMPLEX64, DT_COMPLEX128, DT_BOOL, DT_STRING, 
    DT_QINT8, DT_QINT16, DT_QINT32, DT_QUINT8, DT_QUINT16, DT_INVALID, DataType
)
from hydro_serving_grpc.serving.contract.tensor_pb2 import TensorShape

DTYPE_TO_FIELDNAME = {
    DT_HALF: "half_val",
    DT_FLOAT: "float_val",
    DT_DOUBLE: "double_val",

    DT_INT8: "int_val",
    DT_INT16: "int_val",
    DT_INT32: "int_val",
    DT_INT64: "int64_val",
    DT_UINT8: "int_val",
    DT_UINT16: "int_val",
    DT_UINT32: "uint32_val",
    DT_UINT64: "uint64_val",
    DT_COMPLEX64: "scomplex_val",
    DT_COMPLEX128: "dcomplex_val",
    DT_BOOL: "bool_val",
    DT_STRING: "string_val",
}

PY_TO_DTYPE = {
    int: DT_INT64,
    str: DT_STRING,
    bool: DT_BOOL,
    float: DT_DOUBLE,
    complex: DT_COMPLEX128
}

DTYPE_TO_PY = {v: k for k, v in PY_TO_DTYPE.items()}

ALIAS_TO_DTYPE = {
    "string": DT_STRING,
    "str": DT_STRING,
    "bool": DT_BOOL,

    "float16": DT_HALF,
    "half": DT_HALF,
    "float32": DT_FLOAT,
    "single": DT_FLOAT,
    "float": DT_DOUBLE,  # We treat float the same way Numpy does, look at np.float_
    "float64": DT_DOUBLE,
    "double": DT_DOUBLE,

    "int8": DT_INT8,
    "byte": DT_INT8,
    "int16": DT_INT16,
    "short": DT_INT16,
    "int32": DT_INT32,
    "int64": DT_INT64,
    "int": DT_INT64,

    "uint8": DT_UINT8,
    "ubyte": DT_UINT8,
    "uint16": DT_UINT16,
    "ushort": DT_UINT16,
    "uint32": DT_UINT32,
    "uint64": DT_UINT64,

    "qint8": DT_QINT8,
    "qint16": DT_QINT16,
    "qint32": DT_QINT32,

    "quint8": DT_QUINT8,
    "quint16": DT_QUINT16,

    "complex64": DT_COMPLEX64,
    "complex128": DT_COMPLEX128,
    "complex": DT_COMPLEX128
}

scalar = "scalar"


def alias_to_proto_dtype(name):
    type_ = ALIAS_TO_DTYPE.get(name, DT_INVALID)
    if not type_:
        try:
            type_ = DataType.Value(name)
        except ValueError:
            type_ = DT_INVALID

    return type_


def shape_to_proto(user_shape):
    if isinstance(user_shape, dict):
        user_shape = user_shape.get("dims")

    if user_shape == "scalar":
        shape = TensorShape()
    elif user_shape is None:
        shape = None
    elif isinstance(user_shape, list) or isinstance(user_shape, tuple):
        if not all(isinstance(dim, numbers.Number) for dim in user_shape):
            raise TypeError("shape_list contains incorrect dim", user_shape)
        shape = TensorShape(dims=user_shape)
    else:
        raise ValueError("Invalid shape value", user_shape)
    return shape


# This dict also allows getting proper proto Dtypes for int, str and other builtin python types
NP_TO_HS_DTYPE = {
    np.int8: DT_INT8,
    np.int16: DT_INT16,
    np.int32: DT_INT32,
    np.int64: DT_INT64,
    np.int: DT_INT64,
    np.uint8: DT_UINT8,
    np.uint16: DT_UINT16,
    np.uint32: DT_UINT32,
    np.uint64: DT_UINT64,
    np.float16: DT_HALF,
    np.float32: DT_FLOAT,
    np.float64: DT_DOUBLE,
    np.float: DT_DOUBLE,
    np.float128: None,
    np.complex64: DT_COMPLEX64,
    np.complex128: DT_COMPLEX128,
    np.complex256: None,
    np.bool_: DT_BOOL,
    np.bool: DT_BOOL,
    np.str_: DT_STRING,
    np.unicode_: DT_STRING,
    str: DT_STRING,
}

HS_TO_NP_DTYPE = dict([(v, k) for k, v in NP_TO_HS_DTYPE.items()])


def proto_to_np_dtype(dt):
    if HS_TO_NP_DTYPE.get(dt) is not None:
        return HS_TO_NP_DTYPE[dt]
    else:
        raise TypeError("Datatype {}({}) is not supported in HydroSDK".format(DataType.Name(dt), dt))


def np_to_proto_dtype(dt):
    if NP_TO_HS_DTYPE.get(dt) is not None:
        return NP_TO_HS_DTYPE[dt]
    else:
        raise TypeError("Datatype {} is not supported in HydroSDK".format(dt))


def find_in_list_by_name(some_list: list, name: str):
    """
    Helper created to find ModelField by required name

    :param some_list: list of objects with name field
    :param name: name of object to be found
    :raises ValueError: not found
    :return: object with required name
    """
    for item in some_list:
        if item.name == name:
            return item

    raise ValueError(f"List: {some_list} doesn't have this name: {name}")


class PredictorDT(Enum):
    DICT_PYTHON = "dict"
    DICT_NP_ARRAY = "nparray"
    DF = "dataframe"



import re
from dataclasses import is_dataclass
from typing import Generator, Dict

import grpc
import requests

from hydrosdk.exceptions import BadRequestException, BadResponseException, UnknownException


def grpc_server_on(channel: grpc.Channel, timeout: int) -> bool:
    """
    The channel_ready_future function allows the client to wait for a specified timeout 
    duration (in seconds) for the server to be ready. If our client times out, it raises.

    :param channel:
    :param timeout: timeout value to check for gRPC connection
    :raises grpc.FutureTimeoutError: if server is off
    :return: status bool
    """
    try:
        grpc.channel_ready_future(channel).result(timeout)
        return True
    except grpc.FutureTimeoutError:
        return False


def handle_request_error(resp: requests.Response, error_message: str):
    """
    Handle http response errors.

    :param resp: response from the cluster
    :param error_message: message of the error to be raised
    :raises BadRequestException: for client-side errors
    :raises BadResponseException: for server-side errors
    """
    if resp.ok:
        return None
    if 400 <= resp.status_code < 500:
        raise BadRequestException(error_message)
    if 500 <= resp.status_code < 600:
        raise BadResponseException(error_message)
    else:
        raise UnknownException(error_message)


def read_in_chunks(filename: str, chunk_size: int) -> Generator[bytes, None, None]:
    """
    Generator to read a file peace by peace.
    
    :param filename: name of the file to read
    :param chunk_size: amount of data to read at once
    """
    with open(filename, "rb") as file:
        while True:
            data = file.read(chunk_size)
            if not data:
                break
            yield data


def enable_camel_case(cls):
    """
    Decorator used for translating Python dataclasses with snake_named
     attributes to same-named backend entities with camelCase named attributes.
     If attributes are dataclasses, then they are converted to\from camelCase recursively

     e.g.
     @enable_camel_case
     @dataclass
     class A:
        super_variable: int = 0

    >> A().to_camel_case_dict()
    {"superVariable" : 0}

    >> A.from_camel_case_dict({"superVariable": 23})
    A(super_variable=23)

    :param cls:
    :return:
    """

    def to_camel_case(s):
        components = s.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])

    def to_snake_case(s):
        s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s).lower()

    def from_camel_case_dict(d) -> cls:
        if d is None: return None

        keys = list(d.keys())
        camel_to_snake_keys = dict(((k, to_snake_case(k)) for k in keys))

        snake_case_dict = dict()

        for k, v in d.items():
            snake_case_key = camel_to_snake_keys[k]
            item_type = cls.__annotations__[snake_case_key]
            if is_dataclass(item_type):
                snake_case_dict[snake_case_key] = item_type.from_camel_case_dict(v)
            else:
                try:
                    if item_type._name == 'List' and len(item_type.__args__) == 1 and is_dataclass(item_type.__args__[0]):
                        snake_case_dict[snake_case_key] = [item_type.__args__[0].from_camel_case_dict(x) for x in v]
                        continue
                    else:
                        snake_case_dict[snake_case_key] = v
                except AttributeError:
                    snake_case_dict[snake_case_key] = v
                snake_case_dict[snake_case_key] = v

        return cls(**snake_case_dict)

    def to_camel_case_dict(self) -> Dict:
        camel_cased_dict = dict()
        for k, v in self.__dict__.items():
            if v is None or k.startswith("_"): continue
            camel_case_key = to_camel_case(k)
            if is_dataclass(v):
                camel_cased_dict[camel_case_key] = v.to_camel_case_dict()
            elif isinstance(v, list):
                if all(map(is_dataclass, v)):
                    camel_cased_dict[camel_case_key] = [x.to_camel_case_dict() for x in v]
                else:
                    camel_cased_dict[camel_case_key] = v
            else:
                camel_cased_dict[camel_case_key] = v
        return camel_cased_dict

    setattr(cls, "from_camel_case_dict", from_camel_case_dict)
    setattr(cls, "to_camel_case_dict", to_camel_case_dict)
    return cls



import importlib_metadata

from hydrosdk.modelversion import ModelVersionBuilder, ModelVersion
from hydrosdk.image import DockerImage
from hydrosdk.application import Application
from hydrosdk.cluster import Cluster
from hydrosdk.signature import SignatureBuilder
from hydrosdk.monitoring import MetricSpec, MetricSpecConfig, ThresholdCmpOp
from hydrosdk.deployment_configuration import DeploymentConfiguration, DeploymentConfigurationBuilder


__version__ = importlib_metadata.version("hydrosdk")



from typing import Union, List

from hydrosdk.cluster import Cluster
from hydrosdk.utils import handle_request_error


class ThresholdCmpOp:
    """
    Threshold comparison operator is used to check if ModelVersion is healthy
    Model is healthy if {metric_value}{TresholdCmpOp}{threshold}
    """
    EQ = "Eq"
    NOT_EQ = "NotEq"
    GREATER = "Greater"
    GREATER_EQ = "GreaterEq"
    LESS = "Less"
    LESS_EQ = "LessEq"


class MetricModel:
    """
    Model having extra metric fields
    """

    def __init__(self, modelversion: 'ModelVersion', threshold: float,
                 comparator: ThresholdCmpOp) -> 'MetricModel':
        self.modelversion = modelversion
        self.threshold = threshold
        self.comparator = comparator


class MetricSpecConfig:
    """
    Metric specification config
    """

    def __init__(self, modelversion_id: int, threshold: Union[int, float],
                 threshold_op: ThresholdCmpOp, servable=None) -> 'MetricSpecConfig':
        """
        Create a MetricSpecConfig for the specified ModelVersion.

        :param modelversion_id: an id of the ModelVersion, which will be used as a monitoring metric
        :param threshold: a threshold for the metric
        :param threshold_op: operator to be used to compare metric values against threshold
        """
        self.servable = servable
        self.threshold_op = threshold_op
        self.threshold = threshold
        self.modelversion_id = modelversion_id


class MetricSpec:
    _BASE_URL = "/api/v2/monitoring/metricspec"

    @staticmethod
    def list(cluster: Cluster) -> List['MetricSpec']:
        """
        Send request and returns list with all available metric specs.

        :param cluster: active cluster
        :return: list with all available metric specs
        """
        resp = cluster.request("GET", MetricSpec._BASE_URL)
        handle_request_error(
            resp, f"Failed to list MetricSpecs. {resp.status_code} {resp.text}")
        return [MetricSpec._from_json(cluster, x) for x in resp.json()]

    @staticmethod
    def find_by_id(cluster: Cluster, id: int) -> 'MetricSpec':
        """
        Find MetricSpec by id.

        :param cluster: active cluster
        :param id: 
        :return: MetricSpec
        """
        resp = cluster.request("GET", f"{MetricSpec._BASE_URL}/{id}")
        handle_request_error(
            resp, f"Failed to retrieve MetricSpec for id={id}. {resp.status_code} {resp.text}")
        return MetricSpec._from_json(cluster, resp.json())

    @staticmethod
    def find_by_modelversion(cluster: Cluster, modelversion_id: int) -> List['MetricSpec']:
        """
        Find MetricSpecs by model version.

        :param cluster: active cluster
        :param modelversion_id: ModelVersions for which to return metrics.
        :return: list of MetricSpec objects
        """
        resp = cluster.request("GET", f"{MetricSpec._BASE_URL}/modelversion/{modelversion_id}")
        handle_request_error(
            resp, f"Failed to list MetricSpecs for modelversion_id={modelversion_id}. {resp.status_code} {resp.text}")
        return [MetricSpec._from_json(cluster, x) for x in resp.json()]

    @staticmethod
    def delete(cluster: Cluster, id: int) -> dict:
        """
        Delete MetricSpec.

        :return: result of deletion
        """
        resp = cluster.request("DELETE", f"{MetricSpec._BASE_URL}/{id}")
        handle_request_error(
            resp, f"Failed to delete MetricSpec for id={id}. {resp.status_code} {resp.text}")
        return resp.json()

    @staticmethod
    def create(cluster: Cluster, name: str, modelversion_id: int,
               config: MetricSpecConfig) -> 'MetricSpec':
        """
        Create MetricSpec and returns corresponding instance.

        :param cluster: active cluster
        :param name: name of the metric
        :param modelversion_id: ModelVersion for which to create a MetricSpec
        :param config: MetricSpecConfig, describing MetricSpec
        :return: metricSpec
        """
        metric_spec_json = {
            'name': name,
            'modelVersionId': modelversion_id,
            'config': {
                'modelVersionId': config.modelversion_id,
                'threshold': config.threshold,
                'thresholdCmpOperator': {
                    'kind': config.threshold_op
                }
            }
        }
        resp = cluster.request("POST", MetricSpec._BASE_URL, json=metric_spec_json)
        handle_request_error(
            resp, f"Failed to create a MetricSpec for name={name}, modelversion_id={modelversion_id}. {resp.status_code} {resp.text}")
        return MetricSpec._from_json(cluster, resp.json())

    @staticmethod
    def _from_json(cluster: Cluster, json_dict: dict) -> 'MetricSpec':
        """
        Deserialize MetricSpec from json.

        :param cluster: active cluster
        :param json_dict:
        :return: MetricSpec obj
        """
        return MetricSpec(
            cluster=cluster,
            id=json_dict['id'],
            name=json_dict['name'],
            modelversion_id=json_dict['modelVersionId'],
            config=MetricSpecConfig(
                modelversion_id=json_dict['config']['modelVersionId'],
                threshold=json_dict['config']['threshold'],
                threshold_op=json_dict['config']['thresholdCmpOperator']['kind'],
                servable=json_dict['config'].get('servable')
            ),
        )

    def __init__(self, cluster: Cluster, id: int, name: str, modelversion_id: int,
                 config: MetricSpecConfig) -> 'MetricSpec':
        self.id = id
        self.cluster = cluster
        self.config = config
        self.modelversion_id = modelversion_id
        self.name = name



import json
from typing import Dict
from typing import Optional
from urllib import parse

import grpc
import requests

from hydrosdk.utils import grpc_server_on, handle_request_error
from hydrosdk.exceptions import BadResponseException


class Cluster:
    """
    Cluster responsible for interactions with the server.

    :Example:

    Create a cluster instance only with HTTP connection.

    >>> cluster = Cluster("http-cluster-endpoint")
    >>> print(cluster.build_info())

    Create a cluster instance with both HTTP and gRPC connection.

    >>> from grpc import ssl_channel_credentials
    >>> grpc_credentials = ssl_channel_credentials()
    >>> cluster = Cluster("http-cluster-endpoint", "grpc-cluster-endpoint", ssl=True, grpc_credentials=grpc_credentials)
    >>> print(cluster.build_info())
    """

    def __init__(self, http_address: str, grpc_address: Optional[str] = None,
                 grpc_credentials: Optional[grpc.ChannelCredentials] = None,
                 grpc_options: Optional[list] = None, grpc_compression: Optional[grpc.Compression] = None,
                 timeout: int = 5, check_connection = True, **kwarg) -> 'Cluster':
        """
        A cluster object which hides networking details and provides a connection to a 
        deployed Hydrosphere cluster.

        :param http_address: HTTP endpoint of the cluster
        :param grpc_address: gRPC endpoint of the cluster
        :param grpc_credentials: an optional instance of ChannelCredentials to use for gRPC endpoint
        :param grpc_options: an optional list of key-value pairs to configure the channel
        :param grpc_compression: an optional value indicating the compression method to be
                                 used over the lifetime of the channel
        :param timeout: timeout value to check for gRPC connection
        :returns: Cluster instance
        """

        # TODO: add better url validation (but not python validators lib!)
        parse.urlsplit(http_address)  # check if address is ok
        self.http_address = http_address

        if grpc_address:
            parse.urlsplit(grpc_address)
            self.grpc_address = grpc_address

            if grpc_credentials is not None:
                self.channel = grpc.secure_channel(target=self.grpc_address, credentials=grpc_credentials,
                                                   options=grpc_options,
                                                   compression=grpc_compression)
            else:
                self.channel = grpc.insecure_channel(target=self.grpc_address, options=grpc_options,
                                                     compression=grpc_compression)

            if check_connection:
                if not grpc_server_on(self.channel, timeout):
                    raise ConnectionError(
                        f"Couldn't establish connection with grpc {self.grpc_address}. No connection")

    def request(self, method: str, url: str, **kwargs) -> requests.Response:
        """
        Formats address and url, send request

        :param method: type of request
        :param url: url for a request to be sent to
        :param kwargs: additional args
        :return: request res
        """
        url = parse.urljoin(self.http_address, url)
        return requests.request(method, url, **kwargs)

    def build_info(self) -> Dict[str, str]:
        """
        Returns Manager, Gateway and Sonar services builds information containing version, release commit, etc.

        :return: Dictionary with build information
        """
        manager_bl = self.safe_buildinfo("/api/buildinfo")
        gateway_bl = self.safe_buildinfo("/gateway/buildinfo")
        monitoring_bl = self.safe_buildinfo("/monitoring/buildinfo")
        rootcause_bl = self.safe_buildinfo("/rootcause/buildinfo")
        visualization_bl = self.safe_buildinfo("/visualization/buildinfo")
        stat_bl = self.safe_buildinfo("/stat/buildinfo")
        return {
            "manager": manager_bl,
            "gateway": gateway_bl,
            "monitoring": monitoring_bl,
            "rootcase": rootcause_bl,
            "visualization": visualization_bl,
            "stat": stat_bl,
        }

    def safe_buildinfo(self, url: str) -> Dict[str, str]:
        try:
            resp = self.request("GET", url)
            try:
                handle_request_error(
                    resp, f"Can't fetch buildinfo for {url}. {resp.status_code} {resp.text}")
            except BadResponseException as e:
                return {"status": "Unavailable", "reason": str(e)}
            resp_json = resp.json()
            if 'status' not in resp_json:
                resp_json['status'] = "Ok"
            return resp_json
        except requests.exceptions.ConnectionError as ex:
            return {"status": "Unavailable", "reason": f"Can't establish connection. {ex}"}
        except json.decoder.JSONDecodeError as ex:
            return {"status": "Unknown", "reason": f"Can't parse JSON response. {ex}"}



from docker_image.reference import TaggedReference
from pydantic import BaseModel

class DockerImage(BaseModel):
    name: str
    tag: str
    
    @classmethod
    def from_string(cls, string):
        ref = TaggedReference.parse(string)
        if isinstance(ref, TaggedReference):
            return DockerImage(name = ref['name'], tag = ref['tag'])
        else:
            raise ValueError(f"Couldn't create a DockerImage from the provided image reference: {string}")

    def __str__(self):
        return f"{self.name}:{self.tag}"



"""
This module contains all the code associated with Servables and their management at Hydrosphere platform.
You can learn more about Servables here https://hydrosphere.io/serving-docs/latest/overview/concepts.html#servable.
"""
import re, json, datetime
from enum import Enum
from typing import Dict, List, Optional, Iterable

import sseclient
from sseclient import Event

from hydrosdk import DeploymentConfiguration
from hydrosdk.cluster import Cluster
from hydrosdk.data.types import PredictorDT
from hydrosdk.modelversion import ModelVersion
from hydrosdk.predictor import PredictServiceClient, MonitorableServablePredictionService, UnmonitorableServablePredictionService
from hydrosdk.utils import handle_request_error
from hydrosdk.exceptions import TimeoutException


class ServableStatus(Enum):
    """
    Servable can be in one of four states
    """
    STARTING = 2
    SERVING = 3
    NOT_SERVING = 0
    NOT_AVAILABLE = 1
    UNKNOWN = 4

    @staticmethod
    def from_camel_case(camel_case_servable_status: str) -> 'ServableStatus':
        """
        Get value of ServableStatus enum from CamelCase style string returned from the manager.
        :param camel_case_servable_status:
        :return: ServableStatus

        :Example:
        >>> ServableStatus.from_camel_case("NotServing")
        ServableStatus.NOT_SERVING
        """
        return ServableStatus[re.sub(r'(?<!^)(?=[A-Z])', '_', camel_case_servable_status).upper()]


class Servable:
    """
    Servable is an instance of a model version which is used within applications. Intendend for
    internal usage only.
    """
    _BASE_URL = "/api/v2/servable"

    @staticmethod
    def list(cluster: Cluster) -> List['Servable']:
        """
        Retrieve a list of all servables available at your cluster

        :param cluster: Hydrosphere cluster
        :return: List of all Servables available at your cluster
        """
        resp = cluster.request("GET", Servable._BASE_URL)
        handle_request_error(
            resp, f"Failed to list servables. {resp.status_code} {resp.text}")
        return [Servable._from_json(cluster, servable_json) for
                servable_json in resp.json()]

    @staticmethod
    def find_by_name(cluster: Cluster, servable_name: str) -> 'Servable':
        """
        Finds a serving servable in a cluster

        :param cluster: active cluster
        :param servable_name: a name of the servable
        :raises ServableException:
        :return: Servable
        """

        resp = cluster.request("GET", f"{Servable._BASE_URL}/{servable_name}")
        handle_request_error(
            resp, f"Failed to find servable for name={servable_name}. {resp.status_code} {resp.text}")
        return Servable._from_json(cluster, resp.json())

    @staticmethod
    def create(cluster: Cluster, model_name: str, version: int,
               metadata: Optional[Dict[str, str]] = None,
               deployment_configuration: Optional[DeploymentConfiguration] = None) -> 'Servable':
        """
        Deploy an instance of uploaded model version at your cluster.

        :param deployment_configuration: k8s configurations used to run this servable
        :param cluster: Cluster connected to Hydrosphere
        :param model_name: Name of uploaded model
        :param version: Version of uploaded model
        :param metadata: Information which you can attach to your servable in a form of Dict[str, str]
        :raises ServableException:
        :return: servable
        """
        msg = {
            "modelName": model_name,
            "version": version,
        }
        if metadata:
            msg['metadata'] = metadata

        if deployment_configuration:
            msg['deploymentConfigName'] = deployment_configuration.name

        resp = cluster.request('POST', Servable._BASE_URL, json=msg)
        handle_request_error(resp, f"Failed to create a servable. {resp.status_code} {resp.text}")
        return Servable._from_json(cluster, resp.json())

    @staticmethod
    def delete(cluster: Cluster, servable_name: str) -> dict:
        """
        Shut down and delete servable instance.

        :param cluster: active cluster
        :param servable_name: name of the servable
        :return: json response from serve

        .. warnings also::
          Use with caution. Predictors previously associated with this servable will not be able to connect to it.

        """
        resp = cluster.request("DELETE", f"{Servable._BASE_URL}/{servable_name}")
        handle_request_error(
            resp, f"Failed to delete the servable with name={servable_name}. {resp.status_code} {resp.text}")
        return resp.json()

    @staticmethod
    def _from_json(cluster: Cluster, servable_json: dict) -> 'Servable':
        """
        Deserializes Servable from JSON into a Servable object

        :param cluster: active cluster
        :param servable_json: Servable description in json format
        :return: Servable object
        """
        model_version_id = servable_json['modelVersionId']
        deployment_configuration_name = servable_json.get("deploymentConfigurationName", None)
        status_message = servable_json.get('statusMessage', None)

        return Servable(cluster=cluster,
                        model_version_id=model_version_id,
                        servable_name=servable_json.get('fullName', 'unknown'),
                        status=ServableStatus.from_camel_case(servable_json.get('status', "Unknown")),
                        status_message=status_message,
                        metadata=servable_json.get('metadata'),
                        deployment_configuration_name=deployment_configuration_name)
    
    def __init__(self, cluster: Cluster,
                 model_version_id: int,
                 servable_name: str,
                 status: ServableStatus,
                 status_message: Optional[str],
                 deployment_configuration_name: Optional[str],
                 metadata: Optional[dict] = None) -> 'Servable':
        self.model_version_id = model_version_id
        self.name = servable_name
        self.meta = metadata or {}
        self.cluster = cluster
        self.status = status
        self.status_message = status_message
        self.deployment_configuration_name = deployment_configuration_name
        self.model_version = ModelVersion.find_by_id(cluster, self.model_version_id)

    def logs(self, follow=False) -> Iterable[Event]:
        if follow:
            url = f"{self._BASE_URL}/{self.name}/logs?follow=true"
            resp = self.cluster.request("GET", url, stream=True)
        else:
            url = f"{self._BASE_URL}/{self.name}/logs"
            resp = self.cluster.request("GET", url)
            handle_request_error(
                resp, f"Failed to retrieve logs for {self}. {resp.status_code} {resp.text}")
        return sseclient.SSEClient(resp).events()

    def lock_while_starting(self, timeout: int = 120) -> 'Servable':
        """ Wait for a servable to become ready. """
        updated_servable = Servable.find_by_name(self.cluster, self.name)
        self.status = updated_servable.status
        self.status_message = updated_servable.status_message

        if self.status is ServableStatus.SERVING:
            return self

        deadline_at = datetime.datetime.now().timestamp() + timeout
        events_stream = self.cluster.request("GET", "/api/v2/events", stream=True)
        events_client = sseclient.SSEClient(events_stream)
        try:
            filtered_events = filter(lambda x: x.event == "ServableUpdate", events_client.events())
            servables_stream = map(lambda x: Servable._from_json(self.cluster, json.loads(x.data)), filtered_events)
            target_stream = filter(lambda x: x.name == self.name, servables_stream)

            for servable in target_stream:
                if datetime.datetime.now().timestamp() > deadline_at:
                    raise TimeoutException('Time out waiting for a servable to become ready')
                
                self.status = servable.status
                self.status_message = servable.status_message

                if self.status is ServableStatus.SERVING:
                    return self
        finally:
            events_client.close()
        

    def __str__(self) -> str:
        return f"Servable '{self.name}' for model_version_id {self.model_version_id}"

    def __repr__(self) -> str:
        return f"Servable {self.name}"

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "status": self.status.name,
            "status_message": self.status_message,
            "meta": self.meta,
            "model_version": self.model_version.to_dict(),
            "deployment_configuration": self.deployment_configuration.to_dict(),
        }

    def predictor(self, monitorable=True, return_type=PredictorDT.DICT_NP_ARRAY) -> PredictServiceClient:
        """
        Returns a predictor object which is used to pass data into the deployed Servable

        :param monitorable: If True, the data will be shadowed to the monitoring service
        :param return_type: Specifies into which data format should predictor return Servable outputs
        :return:
        """
        if monitorable:
            impl = MonitorableServablePredictionService(channel=self.cluster.channel, target=self.name)
        else:
            impl = UnmonitorableServablePredictionService(channel=self.cluster.channel, target=self.name)

        return PredictServiceClient(impl=impl, signature=self.model_version.signature,
                                    return_type=return_type)



import numbers
import operator
from enum import Enum
from functools import reduce
from typing import Optional, Union, Iterable, List

import numpy as np
from hydro_serving_grpc.serving.contract.signature_pb2 import ModelSignature
from hydro_serving_grpc.serving.contract.field_pb2 import ModelField
from hydro_serving_grpc.serving.contract.types_pb2 import (
    DT_INVALID, DT_BOOL, DT_FLOAT, DT_HALF, DT_DOUBLE, DT_COMPLEX64, DT_COMPLEX128,
    DT_INT8, DT_INT16, DT_INT32, DT_INT64, DT_UINT8, DT_UINT16, DT_UINT32, DT_UINT64,
    DT_STRING, DataProfileType, DataType,
)

from hydrosdk.data.types import (
    alias_to_proto_dtype, shape_to_proto, PY_TO_DTYPE, np_to_proto_dtype, proto_to_np_dtype,
)
from hydrosdk.exceptions import SignatureViolationException

class ProfilingType(Enum):
    """
    Profiling Types are used to tell monitoring services how to analyse signature fields.
    """
    NONE = 0
    CATEGORICAL = 1
    NOMINAL = 11
    ORDINAL = 12
    NUMERICAL = 2
    CONTINUOUS = 21
    INTERVAL = 22
    RATIO = 23
    IMAGE = 3
    VIDEO = 4
    AUDIO = 5
    TEXT = 6


def field_from_dict(field_name: str, field_dict: dict) -> ModelField:
    """
    Deserialization into ModelField.

    :param field_name:
    :param field_dict: data
    :raises ValueError: If data_dict is invalid
    :return: ModelField
    """

    shape = field_dict.get("shape")

    dtype = field_dict.get("dtype")
    if not dtype:
        dtype = field_dict.get("type")

    subfields = field_dict.get("subfields")
    if not subfields:
        subfields = field_dict.get("fields")

    raw_profile = field_dict.get("profile", "NONE")
    profile = raw_profile.upper()

    if profile not in DataProfileType.keys():
        profile = "NONE"

    result_dtype = None
    result_subfields = None
    if dtype is None:
        if subfields is None:
            raise ValueError("Invalid field. Neither dtype nor subfields are present in dict", field_name, field_dict)
        else:
            subfields_buffer = []
            for k, v in subfields.items():
                subfield = field_from_dict(k, v)
                subfields_buffer.append(subfield)
            result_subfields = subfields_buffer
    else:
        result_dtype = alias_to_proto_dtype(dtype)

    if result_dtype is not None:
        result_field = ModelField(
            name=field_name,
            shape=shape_to_proto(shape),
            dtype=result_dtype,
            profile=profile
        )
    elif result_subfields is not None:
        result_field = ModelField(
            name=field_name,
            shape=shape_to_proto(shape),
            subfields=ModelField.Subfield(data=result_subfields),
            profile=profile
        )
    else:
        raise ValueError("Invalid field. Neither dtype nor subfields are present in dict", field_name, field_dict)
    return result_field


def ModelSignature_to_signature_dict(signature: ModelSignature) -> dict:
    """
    Serializes ModelSignature into a signature dict

    :param signature: model signature obj
    :raises TypeError: If signature invalid
    :return: dict with signature_name, inputs, outputs
    """
    if not isinstance(signature, ModelSignature):
        raise TypeError("signature is not ModelSignature")
    inputs = []
    for i in signature.inputs:
        inputs.append(field_to_dict(i))
    outputs = []
    for o in signature.outputs:
        outputs.append(field_to_dict(o))
    result_dict = {
        "signatureName": signature.signature_name,
        "inputs": inputs,
        "outputs": outputs
    }
    return result_dict


def field_to_dict(field: ModelField) -> dict:
    """
    Serializes model field into name, profile and optional shape

    :param field: model field
    :raises TypeError: If field is invalid
    :return: dict with name and profile
    """
    if not isinstance(field, ModelField):
        raise TypeError("field is not ModelField")
    result_dict = {
        "name": field.name,
        "profile": DataProfileType.Name(field.profile),
        "shape": shape_to_dict(field.shape)
    }

    attach_ds(result_dict, field)
    return result_dict


def attach_ds(result_dict: dict, field) -> dict:
    """
    Adds dtype or subfields

    :param result_dict:
    :param field:
    :raises ValueError: If field invalid
    :return: result_dict with dtype or subfields
    """
    if field.dtype is not None:
        result_dict["dtype"] = DataType.Name(field.dtype)
    elif field.subfields is not None:
        subfields = []
        for f in field.subfields:
            subfields.append(field_to_dict(f))
        result_dict["subfields"] = subfields
    else:
        raise ValueError("Invalid ModelField type")
    return result_dict


def shape_to_dict(shape: Optional['TensorShape']=None) -> dict:
    """
    Serializes model field's shape to dict

    :param shape: TensorShape
    :return: dict with dim
    """
    return None if shape is None else {"dims": list(shape.dims)}


def signature_dict_to_ModelSignature(data: dict) -> ModelSignature:
    """
    Internal method.
    A method that creates ModelSignature out of a signature dict.
    :param data:
    :return:
    """
    signature_name = data.get("signatureName")
    inputs = data.get("inputs")  # list of dicts
    outputs = data.get("outputs")  # list of dicts

    frmt_inputs = []
    for input_ in inputs:

        field_name = input_.pop("name", None)
        if not field_name:
            [(field_name, field_dict)] = input_.items()
        else:
            field_dict = input_

        frmt_inputs.append(field_from_dict(field_name=field_name, field_dict=field_dict))

    frmt_outputs = []
    for output in outputs:
        field_name = output.pop("name", None)
        if not field_name:
            [(field_name, field_dict)] = output.items()
        else:
            field_dict = output

        frmt_outputs.append(field_from_dict(field_name=field_name, field_dict=field_dict))

    return ModelSignature(
        signature_name=signature_name,
        inputs=frmt_inputs,
        outputs=frmt_outputs
    )


def parse_field(name: str, dtype: Union[str, int, np.dtype],
                shape: Iterable[int], profile: ProfilingType = ProfilingType.NONE) -> ModelField:
    """
    Creates a proto ModelField object

    :param name: name of a model field
    :param dtype: data type of model field, either string dtype alias ("double", "int" .. ),
     proto DataType value or name, or a Numpy data type
    :param shape: shape of a model field
    :param profile: profile of a model field
    :raises ValueError: If dtype is invalid
    :return: ModelField proto object
    """
    if dtype is None:
        raise ValueError("Invalid field. Neither dtype nor subfields are present in dict", name)
    elif isinstance(dtype, dict):
        subfields_buffer = []
        for name, v in dtype.items():
            subfield = parse_field(name, v['dtype'], v['shape'], v['profile'])
            subfields_buffer.append(subfield)
        return ModelField(
            name=name,
            shape=shape_to_proto(shape),
            subfields=ModelField.Subfield(data=subfields_buffer),
            profile=profile.value
        )
    else:
        if dtype in DataType.keys():  # exact name e.g. DT_STRING
            result_dtype = dtype
        elif dtype in DataType.values():  # int value of DataType
            result_dtype = dtype
        elif isinstance(dtype, str):  # string alias e.g. 'double'
            result_dtype = alias_to_proto_dtype(dtype)
        elif isinstance(dtype, type):  # type. could be python or numpy type
            result_dtype = PY_TO_DTYPE.get(dtype)
            if not result_dtype:
                result_dtype = np_to_proto_dtype(dtype)
        else:
            result_dtype = DT_INVALID

        if result_dtype == DT_INVALID:
            raise ValueError(f"Invalid contract: {name} field has invalid datatype {dtype}")
        return ModelField(
            name=name,
            shape=shape_to_proto(shape),
            dtype=result_dtype,
            profile=profile.value
        )


class SignatureBuilder:
    def __init__(self, name):
        """
        SignatureBuilder is used to help with the creation of a ModelSignature.

        Example:
            signature = SignatureBuilder('infer') \
                .with_input('x', 'double', 'scalar') \
                .with_output('y', 'double', 'scalar').build()
        """
        self.name = name
        self.inputs = []
        self.outputs = []

    def with_input(self, name: str, dtype: Union[np.dtype, str, type], 
                   shape: Union[str, Iterable[int]], 
                   profile: ProfilingType = ProfilingType.NONE) -> 'SignatureBuilder':
        """
        Adds an input field to the current ModelSignature

        :param name: string containing a name of the field
        :param dtype: type of the field (one of: DataType, numpy's dtype, string representing datatypes
            like 'double', 'int64', standard python types like `int`, `float`, `str`)
        :param shape: shape of the field (one of: 'scalar', iterable of ints)
        :param profile: one of the options from ProfilingType
        :return: SignatureBuilder object
        """
        return self.__with_field(self.inputs, name, dtype, shape, profile)

    def with_output(self, name: str, dtype: Union[np.dtype, str, type], 
                    shape: Union[str, Iterable[int]], 
                    profile: ProfilingType = ProfilingType.NONE) -> 'SignatureBuilder':
        """
        Adds an output field to the current ModelSignature

        :param name: string containing a name of the field
        :param dtype: type of the field (one of: DataType, numpy's dtype, string representing datatypes
            like 'double', 'int64', standard python types like `int`, `float`, `str`)
        :param shape: shape of the field (one of: 'scalar', iterable of ints)
        :param profile: one of the options from ProfilingType
        :return: SignatureBuilder object
        """
        return self.__with_field(self.outputs, name, dtype, shape, profile)

    def build(self) -> ModelSignature:
        """
        Creates ModelSignature
        :return: ModelSignature proto object
        """
        return ModelSignature(
            signature_name=self.name,
            inputs=self.inputs,
            outputs=self.outputs
        )
                    
    def __with_field(self, collection: List[ModelField], name: str, 
                     dtype: Union[np.dtype, str, type], shape: Union[str, Iterable[int]], 
                     profile: ProfilingType = ProfilingType.NONE) -> 'SignatureBuilder':
        """
        Adds fields to the SignatureBuilder

        :param collection: input or output
        :param name: string containing a name of the field
        :param dtype: type of the field (one of: DataType, numpy's dtype, string representing datatypes
            like 'double', 'int64', standard python types like `int`, `float`, `str`)
        :param shape: shape of the field (one of: 'scalar', iterable of ints)
        :param profile: one of the options from ProfilingType
        :return: SignatureBuilder object
        """
        if not isinstance(profile, ProfilingType):
            raise ValueError("`profile` should be of instance ProfilingType.")

        proto_field = parse_field(name, dtype, shape, profile)
        collection.append(proto_field)
        return self


class AnyDimSize(object):
    """
    Validation class for dimensions, used for -1 dims
    """

    def __eq__(self, other):
        """
        If dimension is of Number type than equal

        :param other: dimension
        :raises TypeError: If other not Number
        :return:
        """
        if isinstance(other, numbers.Number):
            return True
        else:
            raise TypeError(f"Unexpected other argument {other}")


def validate_signature(signature: ModelSignature):
    if not signature.signature_name:
        raise SignatureViolationException("Creating model without signature_name is not allowed")
    if len(signature.inputs) == 0:
        raise SignatureViolationException("Creating model without inputs is not allowed")
    if len(signature.outputs) == 0:
        raise SignatureViolationException("Creating model without outputs is not allowed")
    for model_field in signature.inputs:
        if model_field.dtype == 0:
            raise SignatureViolationException("Creating model with invalid dtype in the signature inputs is not allowed")
    for model_field in signature.outputs:
        if model_field.dtype == 0:
            raise SignatureViolationException("Creating model with invalid dtype in the signature outputs is not allowed")


def mock_input_data(signature: ModelSignature):
    """
    Creates dummy input data

    :param signature:
    :return: list of input tensors
    """
    input_tensors = []
    for field in signature.inputs:
        simple_shape = field.shape.dims or [1]
        field_shape = tuple(np.abs(simple_shape))
        size = reduce(operator.mul, field_shape)
        npdtype = proto_to_np_dtype(field.dtype)
        if field.dtype == DT_BOOL:
            x = (np.random.randn(*field_shape) >= 0).astype(np.bool)
        elif field.dtype in [DT_FLOAT, DT_HALF, DT_DOUBLE, DT_COMPLEX128, DT_COMPLEX64]:
            x = np.random.randn(*field_shape).astype(npdtype)
        elif field.dtype in [DT_INT8, DT_INT16, DT_INT32, DT_INT64, DT_UINT8, DT_UINT16, DT_UINT32, DT_UINT64]:
            _min, _max = np.iinfo(npdtype).min, np.iinfo(npdtype).max
            x = np.random.randint(_min, _max, size, dtype=npdtype).reshape(field_shape)
        elif field.dtype == DT_STRING:
            x = np.array(["foo"] * size).reshape(field_shape)
        else:
            raise ValueError(f"{field.dtype} does not support mock data generation yet.")
        input_tensors.append(x)
    return input_tensors



from abc import abstractmethod, ABC
from typing import Union, Dict

import numpy as np
import pandas as pd
from hydro_serving_grpc.serving.runtime.api_pb2 import PredictRequest, PredictResponse
from hydro_serving_grpc.serving.contract.tensor_pb2 import Tensor
from hydro_serving_grpc.serving.gateway.api_pb2 import GatewayPredictRequest
from hydro_serving_grpc.serving.gateway.api_pb2_grpc import GatewayServiceStub
from hydro_serving_grpc.serving.contract.signature_pb2 import ModelSignature

from hydrosdk.data.conversions import convert_inputs_to_tensor_proto, tensor_proto_to_np, tensor_proto_to_py
from hydrosdk.data.types import PredictorDT


class PredictImplementation(ABC):
    @abstractmethod
    def send_data(self, inputs: dict):
        pass


class MonitorableApplicationPredictionService(PredictImplementation):
    """Sends data to an application and shadows it to the monitoring services."""
    def __init__(self, channel, target: str):
        """
        :param channel:
        :param target: Name of an application, which will receive data.
        """
        self.stub = GatewayServiceStub(channel)
        self.target = target

    def send_data(self, inputs: Dict[str, Tensor]) -> PredictResponse:
        request = GatewayPredictRequest(name=self.target, data=inputs)
        return self.stub.PredictApplication(request)


class MonitorableServablePredictionService(PredictImplementation):
    """Sends data to a servable and shadows it to the monitoring services."""
    def __init__(self, channel, target: str):
        """
        :param channel:
        :param target: Name of a servable, which will receive the data.
        """
        self.stub = GatewayServiceStub(channel)
        self.target = target

    def send_data(self, inputs: Dict[str, Tensor]) -> PredictResponse:
        request = GatewayPredictRequest(name=self.target, data=inputs)
        return self.stub.PredictServable(request)


class UnmonitorableServablePredictionService(PredictImplementation):
    """Sends data to a servable without shadowing it to the monitoring services."""
    def __init__(self, channel, target: str):
        """
        :param channel:
        :param target: Name of a servable, which will receive the data.
        """
        self.stub = GatewayServiceStub(channel)
        self.target = target

    def send_data(self, inputs: Dict[str, Tensor]) -> PredictResponse:
        request = GatewayPredictRequest(name=self.target, data=inputs)
        return self.stub.ShadowlessPredictServable(request)


class PredictServiceClient:
    """PredictServiceClient is the main way of passing your data to a deployed ModelVersion"""

    def __init__(self, impl: PredictImplementation, signature: ModelSignature, return_type: PredictorDT):
        """
        Creates a client through which you could send your data.
        :param impl: implementation - either Monitorable (which shadows data to monitoring services), either not
        :param signature: ModelVersion signature which is used to encode/decode your data into proto messages
        :param return_type: One of 3 ways (Pandas, Numpy, Python) to represent a ModelVersion output
        """
        self.impl = impl
        self.signature = signature
        self.return_type = return_type

    def predict(self, inputs: Union[pd.DataFrame, dict, pd.Series]) -> Union[pd.DataFrame, Dict]:
        """
        Sends data to the model version deployed in the cluster and returns the response from it.

        This methods hides the conversion of Python/Numpy objects into a TensorProto objects.

        :param inputs: Input data in either Pandas DataFrame, Numpy or Python lists/scalars.
        :return: Output data in a format specified by `return_type`
        """
        inputs_as_proto = convert_inputs_to_tensor_proto(inputs, self.signature)

        try:
            response = self.impl.send_data(inputs=inputs_as_proto)
            if self.return_type == PredictorDT.DF:
                return self.predict_resp_to_df(response=response)

            elif self.return_type == PredictorDT.DICT_NP_ARRAY:
                return self.predict_resp_to_dict_np(response=response)

            elif self.return_type == PredictorDT.DICT_PYTHON:
                return self.predict_resp_to_dict_pydtype(response=response)
        except Exception as err:
            raise Exception(f"Failed to predict. {str(err)} ")

    @staticmethod
    def predict_resp_to_dict_pydtype(response: PredictResponse) -> Dict:
        """
        Transform PredictResponse into a Dictionary with Python lists/scalars
        :param response: PredictResponse proto message returned from the runtime
        :return: Dictionary with Python list/scalars
        """
        output_tensors_dict = {}
        for tensor_name, tensor_proto in response.outputs.items():
            output_tensors_dict[tensor_name] = tensor_proto_to_py(tensor_proto)
        return output_tensors_dict

    @staticmethod
    def predict_resp_to_dict_np(response: PredictResponse) -> Dict[str, np.array]:
        """
        Transform PredictResponse into a Dictionary with Numpy arrays/scalars
        :param response: PredictResponse proto message returned from the runtime
        :return: Dictionary with Numpy arrays
        """
        output_tensors_dict = dict()
        for tensor_name, tensor_proto in response.outputs.items():
            output_tensors_dict[tensor_name] = tensor_proto_to_np(tensor_proto)
        return output_tensors_dict

    @staticmethod
    def predict_resp_to_df(response: PredictResponse) -> pd.DataFrame:
        """
        Transform PredictResponse into a pandas.DataFrame by using intermediate representation of Dict[str, np.array]
        :param response: PredictResponse proto message returned from the runtime
        :return: pandas DataFrame
        """
        response_dict: Dict[str, np.array] = PredictServiceClient.predict_resp_to_dict_np(response)
        return pd.DataFrame(response_dict)



class HydrosphereException(Exception):
    """ 
    Base exception for SDK library.
    """
    pass


class BadRequestException(HydrosphereException):
    """
    Exception for client-side errors.
    """
    pass


class BadResponseException(HydrosphereException):
    """
`    Exception for server-side errors.
    """
    pass


class TimeoutException(HydrosphereException):
    """
    Exception for timeouts
    """
    pass


class UnknownException(HydrosphereException):
    pass


class ServableException(HydrosphereException):
    """
    Servable class base exception
    """
    pass


class MetricSpecException(HydrosphereException):
    pass


class SignatureViolationException(HydrosphereException):
    """
    Exception raised when signature is violated
    """
    pass



import re
from pydantic import BaseModel
from typing import List, Dict, Optional

from hydrosdk.cluster import Cluster
from hydrosdk.utils import handle_request_error, enable_camel_case
from hydrosdk.builder import AbstractBuilder

class K8SEntity(BaseModel):
    class Config:
        @staticmethod
        def to_camel_case(x: str):
            segments = x.split("_")
            return segments[0] + "".join([x.capitalize() for x in segments[1:]])

        alias_generator = to_camel_case
        allow_population_by_field_name = True

class HorizontalPodAutoScalerSpec(K8SEntity):
    """
    HPA Specification

    :param max_replicas: upper limit for the number of replicas to which the autoscaler can scale up. It cannot be less that minReplicas.
    :param min_replicas: lower limit for the number of replicas to which the autoscaler can scale down.
      It defaults to 1 pod. minReplicas is allowed to be 0 if the alpha feature gate HPAScaleToZero is enabled and at least
      one Object or External metric is configured. Scaling is active as long as at least one metric value is available.
    :param cpu_utilization: target average CPU utilization (represented as a percentage of requested CPU)
      over all the pods; if not specified the default autoscaling policy will be used.
    """
    min_replicas: int
    max_replicas: int
    cpu_utilization: int = None

class NodeSelectorRequirement(K8SEntity):
    """
    A node selector requirement is a selector that contains values, a key, and an operator that relates the key and values.

    :param operator: Represents a key's relationship to a set of values. Valid operators are In, NotIn, Exists, DoesNotExist. Gt, and Lt.
    :param key: The label key that the selector applies to.
    :param values: An array of string values. If the operator is In or NotIn, the values array must be non-empty.
     If the operator is Exists or DoesNotExist, the values array must be empty.
     If the operator is Gt or Lt, the values array must have a single element, which will be interpreted as an integer.
     This array is replaced during a strategic merge patch.
    """
    key: str
    operator: str
    values: List[str] = None

class NodeSelectorTerm(K8SEntity):
    """
    A null or empty node selector term matches no objects. The requirements of them are ANDed.
     The TopologySelectorTerm type implements a subset of the NodeSelectorTerm.
    """
    match_expressions: List[NodeSelectorRequirement] = None
    match_fields: List[NodeSelectorRequirement] = None

class PreferredSchedulingTerm(K8SEntity):
    """
    An empty preferred scheduling term matches all objects with implicit weight 0 (i.e. it's a no-op). A null preferred scheduling
     term matches no objects (i.e. is also a no-op).

    :param preference: A node selector term, associated with the corresponding weight.
    :param weight: Weight associated with matching the corresponding nodeSelectorTerm, in the range 1-100.
    """
    weight: int
    preference: NodeSelectorTerm

class NodeSelector(K8SEntity):
    """
    A node selector represents the union of the results of one or more label queries over a set of nodes; that is,
     it represents the OR of the selectors represented by the node selector terms.
    """
    node_selector_terms: List[NodeSelectorTerm] = []

class LabelSelectorRequirement(K8SEntity):
    """
    A label selector requirement is a selector that contains values, a key, and an operator that relates the key and values.

    :param key: key is the label key that the selector applies to.
    :param operator: operator represents a key's relationship to a set of values. Valid operators are In, NotIn, Exists and DoesNotExist.
    :param values: values is an array of string values. If the operator is In or NotIn, the values array must be non-empty.
     If the operator is Exists or DoesNotExist, the values array must be empty. This array is replaced during a strategic merge patch.
    """
    key: str
    operator: str
    values: Optional[List[str]] = None

class LabelSelector(K8SEntity):
    """
    A label selector is a label query over a set of resources. The result of matchLabels and matchExpressions are ANDed.
    An empty label selector matches all objects. A null label selector matches no objects.

    :param match_labels: map of {key,value} pairs. A single {key,value} in the matchLabels map is equivalent
      to an element of matchExpressions, whose key field is "key", the operator is "In", and the values array contains only "value".
      The requirements are ANDed.
    :param match_expressions: matchExpressions is a list of label selector requirements. The requirements are ANDed.
    """
    match_expressions: Optional[List[LabelSelectorRequirement]] = None
    match_labels: Optional[Dict[str, str]] = None

class PodAffinityTerm(K8SEntity):
    """
    Defines a set of pods (namely those matching the labelSelector relative to the given namespace(s)) that this pod should be co-located
    (affinity) or not co-located (anti-affinity) with, where co-located is defined as running on a node whose value of the label with key
    <topologyKey> matches that of any node on which a pod of the set of pods is running

    :param label_selector: A label query over a set of resources, in this case pods.
    :param namespaces: namespaces specifies which namespaces the labelSelector applies to (matches against);
     null or empty list means "this pod's namespace"
    :param topology_key: This pod should be co-located (affinity) or not co-located (anti-affinity) with the pods
     matching the labelSelector in the specified namespaces, where co-located is defined as running on a node whose
     value of the label with key topologyKey matches that of any node on which any of the selected pods is running.
     Empty topologyKey is not allowed.
    """
    label_selector: LabelSelector
    topology_key: str
    namespaces: Optional[List[str]] = None

class WeightedPodAffinityTerm(K8SEntity):
    """
    The weights of all of the matched WeightedPodAffinityTerm fields are added per-node to find the most preferred node(s)

    :param pod_affinity_term: Required. A pod affinity term, associated with the corresponding weight.
    :param weight: weight associated with matching the corresponding podAffinityTerm, in the range 1-100.
    """
    weight: int
    pod_affinity_term: PodAffinityTerm

class NodeAffinity(K8SEntity):
    """
    Group of node affinity scheduling rules.
    """
    preferred_during_scheduling_ignored_during_execution: List[PreferredSchedulingTerm] = []
    required_during_scheduling_ignored_during_execution: Optional[NodeSelector] = None

class PodAffinity(K8SEntity):
    """
    Pod affinity is a group of inter pod affinity scheduling rules.
    """
    preferred_during_scheduling_ignored_during_execution: List[WeightedPodAffinityTerm] = []
    required_during_scheduling_ignored_during_execution: List[PodAffinityTerm] = []

class PodAntiAffinity(K8SEntity):
    """
    Pod anti affinity is a group of inter pod anti affinity scheduling rules.
    """
    preferred_during_scheduling_ignored_during_execution: List[WeightedPodAffinityTerm] = []
    required_during_scheduling_ignored_during_execution: List[PodAffinityTerm] = []

class Affinity(K8SEntity):
    """
    Group of affinity scheduling rules.

    :param node_affinity: node affinity scheduling rules for the pod.
    :param pod_affinity: pod affinity scheduling rules (e.g. co-locate this pod in the same node, zone, etc. as some other pod(s)).
    :param pod_anti_affinity: pod anti-affinity scheduling rules
        (e.g. avoid putting this pod in the same node, zone, etc. as some other pod(s)).
    """
    node_affinity: Optional[NodeAffinity] = None
    pod_affinity: Optional[PodAffinity] = None
    pod_anti_affinity: Optional[PodAntiAffinity] = None

class Toleration(K8SEntity):
    """
    The pod this Toleration is attached to tolerates any taint that matches the triple <key,value,effect> using the matching operator <operator>.

    :param effect: Effect indicates the taint effect to match.
        Empty means match all taint effects. When specified, allowed values are NoSchedule, PreferNoSchedule and NoExecute.
    :param key: Key is the taint key that the toleration applies to. Empty means match all taint keys.
     If the key is empty, operator must be Exists; this combination means to match all values and all keys.
    :param operator: Operator represents a key's relationship to the value. Valid operators are Exists and Equal. Defaults to Equal.
     Exists is equivalent to wildcard for value, so that a pod can tolerate all taints of a particular category.
    :param toleration_seconds: TolerationSeconds represents the period of time the toleration (which must be of effect NoExecute,
     otherwise this field is ignored) tolerates the taint. By default, it is not set, which means tolerate the taint forever (do not evict).
     Zero and negative values will be treated as 0 (evict immediately) by the system.
    :param value: Value is the taint value the toleration matches to. If the operator is Exists, the value should be empty,
     otherwise just a regular string.
    """
    operator: str
    toleration_seconds: Optional[int] = None
    key: Optional[str] = None
    value: Optional[str] = None
    effect: Optional[str] = None

class PodSpec(K8SEntity):
    """
    PodSpec contains all configurations related to a pod.
    """
    node_selector: Dict[str, str] = {}
    tolerations: List[Toleration] = []
    affinity: Optional[Affinity] = None

class DeploymentSpec(K8SEntity):
    """
    DeploymentSpec contains all configurations related to a deployment.
    """
    replica_count: int = 1

class ResourceRequirements(K8SEntity):
    """
    Resources required by the container.

    :param limits: the maximum amount of compute resources allowed
    :param requests: Requests describes the minimum amount of compute resources required.
                     If Requests is omitted for a container, it defaults to Limits if that is explicitly specified,
                     otherwise to an implementation-defined value.
    """
    limits: Dict[str, str]
    requests: Dict[str, str] = None

    def __post_init__(self):
        quantity_validation_regex = r"^([+-]?[0-9.]+)([eEinumkKMGTP]*[-+]?[0-9]*)$"
        for resource, quantity in self.limits.items():
            if not re.match(quantity_validation_regex, quantity):
                raise ValueError(f"{resource} limit ({quantity}) is invalid. Quantity must match '{quantity_validation_regex}'")
        for resource, quantity in self.requests.items():
            if not re.match(quantity_validation_regex, quantity):
                raise ValueError(f"{resource} request ({quantity}) is invalid. Quantity must match '{quantity_validation_regex}'")

class ContainerSpec(K8SEntity):
    """
    ContainerSpec contains all configurations related to a container.
    """
    resources: Optional[ResourceRequirements] = None
    env: Optional[Dict[str, str]] = None

class DeploymentConfiguration(K8SEntity):
    """
    DeploymentConfiguration encapsulates k8s configs about how your Servables and ModelVariants should run.

    :param name: Unique Name of a Deployment Configuration
    :param hpa: HorizontalPodAutoScaler specification
    :param pod: Pod specification
    :param container: Container specification
    :param deployment: Deployment specification
    """
    name: str
    hpa: Optional[HorizontalPodAutoScalerSpec] = None
    pod: Optional[PodSpec] = None
    container: Optional[ContainerSpec] = None
    deployment: Optional[DeploymentSpec] = None

    _BASE_URL: str = "/api/v2/deployment_configuration"

    @staticmethod
    def list(cluster: Cluster) -> List['DeploymentConfiguration']:
        """
        List all deployment configurations

        :param cluster:
        :return:
        """
        resp = cluster.request("GET", DeploymentConfiguration._BASE_URL)
        handle_request_error(resp, f"Failed to get a list of Deployment Configurations - {resp.status_code} {resp.text}")
        return [DeploymentConfiguration.parse_obj(app_json) for app_json in resp.json()]

    @staticmethod
    def find(cluster: Cluster, name: str) -> 'DeploymentConfiguration':
        """
        Search a deployment configuration by name

        :param cluster:
        :param name:
        :return:
        """
        resp = cluster.request("GET", f"{DeploymentConfiguration._BASE_URL}/{name}")
        handle_request_error(resp, f"Failed to find Deployment Configuration named {name} {resp.status_code} {resp.text}")
        return DeploymentConfiguration.parse_obj(resp.json())

    @staticmethod
    def delete(cluster: Cluster, name: str) -> dict:
        """
        Delete deployment configuration from Hydrosphere cluster

        :param cluster: Active Cluster
        :param name: Deployment configuration name
        :return: json response from the server
        """
        resp = cluster.request("DELETE", f"{DeploymentConfiguration._BASE_URL}/{name}")
        handle_request_error(resp, f"Failed to delete Deployment Configuration named {name}: {resp.status_code} {resp.text}")
        return resp
    
    def to_dict(self):
        return self.dict(by_alias=True)


class DeploymentConfigurationBuilder(AbstractBuilder):
    """
    DeploymentConfigBuilder is used to create a new DeploymentConfiguration.

    :Example:

    Create a new deployment configuration.

    >>> from hydrosdk import Cluster, DeploymentConfigurationBuilder
    >>> cluster = Cluster('http-cluster-endpoint')
    >>> config_builder = DeploymentConfigurationBuilder(name="new_config")
    >>> config = config_builder.with_hpa(max_replicas=10, min_replicas=2, target_cpu_utilization_percentage=80). \
                    with_pod_node_selector({"label1": "key1", "foo": "bar"}). \
                    with_resource_requirements(limits={"cpu": "4", "memory": "4G"}, \
                                               requests={"cpu": "2", "memory": "2G"}). \
                    with_env({"ENVIRONMENT": "1"}).
                    with_replicas(replica_count=4). \
                    with_toleration(effect="PreferNoSchedule", key="equalToleration", toleration_seconds=30, operator="Exists").\
                    build(cluster)
    """

    def __init__(self, name: str) -> "DeploymentConfigurationBuilder":
        """
        Create new Deployment Config Builder

        :param name: Name of created deployment  configuration
        """
        self.name = name
        self.hpa = None
        self.pod_spec = None
        self.container_spec = None
        self.deployment_spec = None

    def with_hpa(self, max_replicas, min_replicas=1, target_cpu_utilization_percentage=None) -> 'DeploymentConfigurationBuilder':
        """
        Adds a HorizontalPodAutoScaler specs to a DeploymentConfiguration.

        :param min_replicas: min_replicas is the lower limit for the number of replicas to which the autoscaler can scale down.
                             It defaults to 1 pod. minReplicas is allowed to be 0 if the alpha feature gate HPAScaleToZero
                             is enabled and at least one Object or External metric is configured. Scaling is active as long
                             as at least one metric value is available.
        :param max_replicas: upper limit for the number of pods that can be set by the autoscaler; cannot be smaller than MinReplicas.
        :param target_cpu_utilization_percentage: target average CPU utilization (represented as a percentage of requested CPU)
         over all the pods; if not specified the default autoscaling policy will be used.
        """
        if self.hpa is not None:
            raise ValueError("HorizontalPodAutoScalerSpec is already set")
        if max_replicas < min_replicas:
            raise ValueError(f"max_replicas ({max_replicas}) cannot be smaller than min_replicas ({min_replicas}).")
        self.hpa = HorizontalPodAutoScalerSpec(min_replicas=min_replicas, max_replicas=max_replicas, cpu_utilization=target_cpu_utilization_percentage)
        return self

    def with_pod_node_selector(self, node_selector: Dict[str, str]) -> 'DeploymentConfigurationBuilder':
        """
        Adds selector which must be true for the pod to fit on a node.

        :param node_selector: Selector which must match a node's labels for the pod to be scheduled on that node
        :return:
        """
        if self.pod_spec is None:
            self.pod_spec = PodSpec(node_selector=node_selector)
        elif self.pod_spec.node_selector is not None:
            raise ValueError(f"Cannot set Node Selector as {node_selector},"
                             f" it is already set to {self.pod_spec.node_selector}")
        else:
            self.pod_spec.node_selector = node_selector
        return self

    def with_affinity(self, affinity: Affinity) -> 'DeploymentConfigurationBuilder':
        """
        Adds pod scheduling constraints

        :param affinity: Group of affinity scheduling rules
        :return:
        """
        if self.pod_spec is None:
            self.pod_spec = PodSpec(affinity=affinity)
        elif self.pod_spec.affinity is not None:
            raise ValueError("Affinity is already set")
        else:
            self.pod_spec.affinity = affinity
        return self

    def with_resource_requirements(self, limits: Dict[str, str], requests: Dict[str, str] = None) -> 'DeploymentConfigurationBuilder':
        """
        Specify resources required by this container.

        :param limits: the maximum amount of compute resources allowed
        :param requests: Requests describes the minimum amount of compute resources required.
                         If Requests is omitted for a container, it defaults to Limits if that is explicitly specified,
                         otherwise to an implementation-defined value.
        :return:
        """
        new_resource_requirements = ResourceRequirements(limits=limits, requests=requests)
        if self.container_spec is None:
            self.container_spec = ContainerSpec(resources=new_resource_requirements)
        elif self.container_spec.resources is not None:
            raise ValueError(f"Cannot set Resource Requirements as {new_resource_requirements},"
                             f" it is already set to {self.container_spec.resources}")
        else:
            self.container_spec.resources = new_resource_requirements
        return self

    def with_env(self, env: Dict[str, str]) -> 'DeploymentConfigurationBuilder':
        """
        Specify env for the container.

        :param env: dictionary with the required environment variables.
        """
        if self.container_spec is None:
            self.container_spec = ContainerSpec(env=env)
        elif self.container_spec.env is not None:
            raise ValueError(f"Cannot set env as {env},"
                             f" they are already set to {self.container_spec.env}")
        else:
            self.container_spec.env = env
        return self
        

    def with_toleration(self,
                        effect: str = None,
                        key: str = None,
                        operator: str = "Equal",
                        toleration_seconds: int = None,
                        value: str = None) -> 'DeploymentConfigurationBuilder':
        """
        Specify pod's toleration. The pod this Toleration is attached to tolerates any taint that matches the triple <key,value,effect>
         using the matching operator <operator>.

        :param effect: Effect indicates the taint effect to match.
            Empty means match all taint effects. When specified, allowed values are NoSchedule, PreferNoSchedule and NoExecute.
        :param key: Key is the taint key that the toleration applies to. Empty means match all taint keys.
         If the key is empty, operator must be Exists; this combination means to match all values and all keys.
        :param operator: Operator represents a key's relationship to the value. Valid operators are Exists and Equal. Defaults to Equal.
         Exists is equivalent to wildcard for value, so that a pod can tolerate all taints of a particular category.
        :param toleration_seconds: TolerationSeconds represents the period of time the toleration (which must be of effect NoExecute,
         otherwise this field is ignored) tolerates the taint. By default, it is not set, which means tolerate the taint forever (do not evict).
         Zero and negative values will be treated as 0 (evict immediately) by the system.
        :param value: Value is the taint value the toleration matches to. If the operator is Exists, the value should be empty,
         otherwise just a regular string.
        """
        new_toleration = Toleration(effect=effect, key=key, operator=operator, toleration_seconds=toleration_seconds, value=value)
        if self.pod_spec is None:
            self.pod_spec = PodSpec(tolerations=[new_toleration])
        else:
            self.pod_spec.tolerations.append(new_toleration)
        return self

    def with_replicas(self, replica_count=1) -> 'DeploymentConfigurationBuilder':
        """
        Specify number of desired pods for a deployment

        :param replica_count: Number of desired pods. This is a pointer to distinguish between explicit zero and not specified.
         Defaults to 1.
        :return:
        """
        if self.deployment_spec is None:
            self.deployment_spec = DeploymentSpec(replica_count=replica_count)
        elif self.deployment_spec.replica_count is not None:
            raise ValueError("Replica Count is already set")
        else:
            self.deployment_spec.replica_count = replica_count
        return self
    
    def _with_hpa_spec(self, hpa: Optional[HorizontalPodAutoScalerSpec] = None) -> 'DeploymentConfigurationBuilder':
        """
        Add a HorizontalPodAutoScalerSpec.
        """
        if hpa is None:
            return self
        if not isinstance(hpa, HorizontalPodAutoScalerSpec):
            raise ValueError("hpa should be of type HorizontalPodAutoScalerSpec")
        if self.hpa is not None:
            raise ValueError("HorizontalPodAutoScalerSpec is already set")
        self.hpa = hpa
        return self
    
    def _with_deployment_spec(self, deployment: Optional[DeploymentSpec] = None) -> 'DeploymentConfigurationBuilder':
        """
        Add a DeploymentSpec.
        """
        if deployment is None:
            return self
        if not isinstance(deployment, DeploymentSpec):
            raise ValueError("deployment should be of type HorizontalPodAutoScalerSpec")
        if self.deployment_spec is not None:
            raise ValueError("DeploymentSpec is already set")
        self.deployment_spec = deployment
        return self

    def _with_container_spec(self, container: Optional[ContainerSpec]) -> 'DeploymentConfigurationBuilder':
        """
        Add a ContainerSpec.
        """
        if container is None:
            return self
        if not isinstance(container, ContainerSpec):
            raise ValueError("container should be of type ContainerSpec")
        if self.container_spec is not None:
            raise ValueError("ContainerSpec is already set")
        self.container_spec = container
        return self
    
    def _with_pod_spec(self, pod: Optional[PodSpec]) -> 'DeploymentConfigurationBuilder':
        """
        Add a PodSpec.
        """
        if pod is None:
            return self
        if not isinstance(pod, PodSpec):
            raise ValueError("pod should be of type PodSpec")
        if self.pod_spec is not None:
            raise ValueError("PodSpec is already set")
        self.pod_spec = pod
        return self
    
    def build(self, cluster: Cluster) -> DeploymentConfiguration:
        """
        Create the Deployment Configuration in your Hydrosphere cluster.

        :return: Deployment Configuration object from the Hydrosphere cluster
        """
        config = DeploymentConfiguration(
            name=self.name,
            hpa=self.hpa,
            pod=self.pod_spec,
            container=self.container_spec,
            deployment=self.deployment_spec
        )
        resp = cluster.request("POST", DeploymentConfiguration._BASE_URL, json=config.to_dict())
        handle_request_error(resp, f"Failed to upload new Deployment Configuration. {resp.status_code} {resp.text}")
        return DeploymentConfiguration.parse_obj(resp.json())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hydrosdk', 'hydrosdk.data']

package_data = \
{'': ['*']}

install_requires = \
['docker-image-py>=0.1.10,<0.2.0',
 'hydro-serving-grpc==3.0.0',
 'importlib_metadata>=1.7.0,<2.0.0',
 'numpy==1.18.3',
 'pandas>=1.1.5',
 'pydantic>=1.8.2,<2.0.0',
 'pyyaml>=5.4.1,<6.0.0',
 'requests==2.23.0',
 'requests_toolbelt>=0.9.1,<0.10.0',
 'sseclient-py>=1.7,<2.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses==0.7']}

setup_kwargs = {
    'name': 'hydrosdk',
    'version': '3.0.1a0',
    'description': " This package's purpose is to provide a simple and convenient way of integrating user's workflow scripts with Serving API.",
    'long_description': "# Hydrosphere Serving SDK\n\n\n[![License](https://img.shields.io/badge/License-Apache%202.0-green.svg)](https://opensource.org/licenses/Apache-2.0)\n[![PyPI version](https://badge.fury.io/py/hydrosdk.svg)](https://badge.fury.io/py/hydrosdk)\n\nThe package contains an implementation of [Hydroserving](https://github.com/Hydrospheredata/hydro-serving) API.\n\nThis package's purpose is to provide a simple and convenient way\nof integrating user's workflow scripts with Serving API.\n\nRead the full documentation [here](https://hydrospheredata.github.io/hydro-serving-sdk/).\n\n## Installation\n```\npip install hydrosdk\n```\n\n## Testing\nTested on python 3.6, 3.7, 3.8.\n\n1. `poetry install`\n2. `poetry run tox`\n\n\n",
    'author': 'Hydrospheredata',
    'author_email': None,
    'maintainer': 'Bulat Lutfullin',
    'maintainer_email': 'blutfullin@provectus.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



import datetime
import json
import logging
import os
import tarfile
import time
from enum import Enum
from typing import Optional, Dict, List, Tuple, Iterator
from hydro_serving_grpc.serving.runtime.api_pb2 import PredictRequest, PredictResponse

import sseclient
import requests
from sseclient import Event
from hydro_serving_grpc.serving.contract.signature_pb2 import ModelSignature
from hydro_serving_grpc.serving.manager.entities_pb2 import (
    ModelVersion as ModelVersionProto, DockerImage as DockerImageProto,
)
from requests_toolbelt.multipart.encoder import MultipartEncoder

from hydrosdk.cluster import Cluster
from hydrosdk.signature import ModelSignature_to_signature_dict, validate_signature, signature_dict_to_ModelSignature
from hydrosdk.image import DockerImage
from hydrosdk.monitoring import MetricSpec, MetricSpecConfig, MetricModel, ThresholdCmpOp
from hydrosdk.exceptions import HydrosphereException, TimeoutException, BadResponseException, BadRequestException
from hydrosdk.utils import handle_request_error, read_in_chunks
from hydrosdk.builder import AbstractBuilder
from hydro_serving_grpc.monitoring.sonar.api_pb2_grpc import MonitoringServiceStub
from hydro_serving_grpc.monitoring.sonar.entities_pb2 import ExecutionInformation, ExecutionMetadata


def _upload_training_data(cluster: Cluster, modelversion_id: int, path: str) -> 'DataUploadResponse':
    """
    Upload training data to Hydrosphere

    :param cluster: Cluster instance
    :param modelversion_id: Id of the model version, for which to upload training data
    :param path: Path to the training data
    :raises BadResponseException: if request failed to process by Hydrosphere
    :return: DataUploadResponse obj
    """
    if path.startswith('s3://'):
        resp = _upload_s3_file(cluster, modelversion_id, path)
    else:
        resp = _upload_local_file(cluster, modelversion_id, path)
    if resp.ok:
        return DataUploadResponse(cluster, modelversion_id)
    raise BadResponseException(f'Failed to upload training data: {resp.text}')


def _upload_local_file(cluster: Cluster, modelversion_id: int, path: str, 
                       chunk_size=1024) -> requests.Response:
    """
    Internal method for uploading local training data to Hydrosphere.

    :param cluster: active cluster
    :param modelversion_id: modelversion_id for which to upload training data
    :param path: path to a local file
    :param chunk_size: chunk size to use for streaming
    """    
    gen = read_in_chunks(path, chunk_size)
    url = f'/monitoring/profiles/batch/{modelversion_id}'
    return cluster.request("POST", url, data=gen, stream=True)


def _upload_s3_file(cluster: Cluster, modelversion_id: int, path: str) -> requests.Response:
    """
    Internal method for submitting training data from S3 to Hydrosphere.

    :param cluster: Cluster instance
    :param url: url to which submit S3 path
    :param path: S3 path to training data
    """
    url = f'/monitoring/profiles/batch/{modelversion_id}/s3'
    return cluster.request("POST", url, json={"path": path})
    

def resolve_paths(path: str, payload: List[str]) -> Dict[str, str]:
    """
    Appends each element of payload to the path and makes {resolved_path: payload_element} dict

    :param path: absolute path
    :param payload: list of relative paths
    :return: dict with {resolved_path: payload_element}
    """
    return {os.path.normpath(os.path.join(path, v)): v for v in payload}


def _analyze(model, stub, request_id, request, response=None, error=None):
    if stub is None:
        raise HydrosphereException("Can't use this method without GRPC cluster")
    metadata = ExecutionMetadata(
        request_id = request_id,
        model_version_id = model.id,
        model_name = model.name,
        model_version = model.version,
        signature_name = model.signature.signature_name
    )
    if response is not None:
        ei = ExecutionInformation(
            request = request,
            metadata = metadata,
            response = response
        )
    elif error is not None:
        ei = ExecutionInformation(
            request = request,
            metadata = metadata,
            error = error
        )
    else:
        raise ValueError("Need to specify either response or error")
    stub.Analyze(ei)
    return ei 


class MonitoringConfiguration:
    def __init__(self, batch_size: int):
        self.batch_size = batch_size

    def to_dict(self):
        return {
            "batchSize": self.batch_size
        }
    

class ModelVersionBuilder(AbstractBuilder):
    """
    A local instance of the model yet to be uploaded to the cluster.
    (https://hydrosphere.io/serving-docs/latest/overview/concepts.html#models)

    :Example:

    Create a local instance of the model and upload it to the cluster.

    >>> from hydrosdk.cluster import Cluster
    >>> from hydrosdk.image import DockerImage
    >>> from hydrosdk.signature import SignatureBuilder, ProfilingType
    >>> cluster = Cluster("http-cluster-endpoint")
    >>> builder = ModelVersionBuilder("my-model", "/path/to/my-model")
    >>> builder.with_runtime(DockerImage("hydrosphere/serving-runtime-python-3.7", tag="latest"))
    >>> builder.with_payload(["src/func_main.py", "requirements.txt"])
    >>> signature = SignatureBuilder("predict") \
            .with_input("x", int, "scalar", ProfilingType.NUMERICAL) \
            .with_output("y", float, "scalar", ProfilingType.NUMERICAL) \
            .build()
    >>> builder.with_signature(signature)
    >>> builder.with_install_command("pip install -r requirements.txt")
    >>> builder.with_training_data("training-data.csv")
    >>> model_version = builder.build(cluster)
    >>> model_version.lock_till_released(timeout=120)
    >>> data_upload_response = model_version.upload_training_data()
    """
    def __init__(self, name: str, path: str) -> 'ModelVersionBuilder':
        """
        :param name: a name of the model
        :param path: a path to the root folder of the model
        """
        self.name = name
        self.path = path
        self.runtime = None
        self.signature = None
        self.payload = {}
        self.metadata = {}
        self.install_command = None
        self.training_data = None
        self.monitoring_configuration = None

    def with_runtime(self, runtime: DockerImage) -> 'ModelVersionBuilder':
        """
        Specify runtime for the model version. 

        :param runtime: a docker image used to run model version's code.
        """
        if not isinstance(runtime, DockerImage):
            raise TypeError("runtime is not of type DockerImage")
        self.runtime = runtime
        return self
    
    def with_signature(self, signature: ModelSignature) -> 'ModelVersionBuilder':
        """
        Specify signature for the model version.

        :param signature: ModelSignature, which specifies the name of the inference function, 
                          its types and shapes of both inputs and outputs.
        """
        if not isinstance(signature, ModelSignature):
            raise TypeError("signature is not of type ModelSignature")
        validate_signature(signature)
        self.signature = signature
        return self
    
    def with_payload(self, payload: List[str]) -> 'ModelVersionBuilder':
        """
        Specify payload for the model version.

        :param payload: a list of paths to files (absolute or relative) with any 
                        additional resources that will be exported to the container.
        """
        self.payload = resolve_paths(path=self.path, payload=payload)
        return self
    
    def with_metadata(self, metadata: Dict[str, str]) -> 'ModelVersionBuilder':
        """
        Specify metadata for the model version.

        :param metadata: a metadata to describe the model version.
        """
        if not isinstance(metadata, dict):
            raise TypeError("metadata is not a dict")
        for key, value in metadata.items():
            if not isinstance(key, str):
                raise TypeError(str(key) + " key from metadata is not a string")
            if not isinstance(value, str):
                raise TypeError(str(value) + " value from metadata is not a string")
        self.metadata = metadata
        return self
    
    def with_install_command(self, install_command: str) -> 'ModelVersionBuilder':
        """
        Specify install command for the model.

        :param install_command: a command to run within a runtime to prepare a model 
                                version environment.
        """
        if install_command and not isinstance(install_command, str):
            raise TypeError("install_command should be a string")
        self.install_command = install_command
        return self
    
    def with_training_data(self, training_data: str) -> 'ModelVersionBuilder':
        """
        Specify training data for the model version.
        
        :param training_data: path (absolute, relative or an S3 URI) to a csv file with 
                              the training data 
        """
        if not isinstance(training_data, str):
            raise TypeError("training_data should be a path presented as a string")
        if training_data.startswith("s3://"): # pass S3 paths as is
            self.training_data = training_data
        else:  # assume that it's local file
            if not os.path.isabs(training_data):
                full_training_path = os.path.normpath(os.path.join(self.path, training_data))
            else:
                full_training_path = training_data # absolute paths as is
            if not os.path.exists(full_training_path):
                raise FileNotFoundError(f"Can't find training data file {full_training_path}")
            self.training_data = full_training_path

        return self

    def with_monitoring_configuration(self, monitoring_configuration: MonitoringConfiguration) -> 'ModelVersionBuilder':
        """
        Specify monitoring configuration for the model version.

        :param monitoring_configuration: Specifies a configuration to be used when monitoring 
                                         the model version.
        """
        if not isinstance(monitoring_configuration, MonitoringConfiguration):
            raise TypeError("monitoring_configuration should be of type MonitoringConfiguration")
        self.monitoring_configuration = monitoring_configuration
        return self
    
    def __repr__(self):
        return f"ModelVersionBuilder(name={self.name})"

    def __create_tarball(self) -> str:
        logger = logging.getLogger("ModelDeploy")
        hs_folder = ".hs"
        os.makedirs(hs_folder, exist_ok=True)
        tarpath = os.path.join(hs_folder, f"{self.name}-{datetime.datetime.now()}")

        logger.debug("Creating payload tarball %s for %s model", tarpath, self.name)
        with tarfile.open(tarpath, "w:gz") as tar:
            for source, target in self.payload.items():
                logger.debug("Archiving %s as %s", source, target)
                tar.add(source, arcname=target)
        return tarpath
                
    def build(self, cluster: Cluster) -> 'ModelVersion':
        """
        Upload local model instance to the cluster.

        :param cluster: active cluster
        :return: ModelVersion object
        """
        if self.runtime is None:
            raise ValueError("runtime is not specified for the model")
        if self.signature is None:
            raise ValueError("signature is not specified for the model")

        meta = {
            "name": self.name,
            "runtime": self.runtime.dict(),
            "modelSignature": ModelSignature_to_signature_dict(self.signature),
            "installCommand": self.install_command,
            "metadata": self.metadata
        }
        
        if self.monitoring_configuration:
            meta.update({"monitoringConfiguration": self.monitoring_configuration.to_dict()})

        tarpath = self.__create_tarball()
        encoder = MultipartEncoder(
            fields={
                "payload": ("filename", open(tarpath, "rb")),
                "metadata": json.dumps(meta)
            }
        )
        
        resp = cluster.request("POST", "/api/v2/model/upload", 
            data=encoder, headers={'Content-Type': encoder.content_type})
        handle_request_error(
            resp, f"Failed to upload local model. {resp.status_code} {resp.text}")

        modelversion = ModelVersion._from_json(cluster, resp.json())
        modelversion.training_data = self.training_data
        return modelversion

    def build_external(self, cluster: Cluster) -> 'ModelVersion':
        """
        Create an external ModelVersion on the cluster. 

        :param cluster: active cluster
        :return: ModelVersion object
        """
        if self.signature is None:
            raise ValueError("signature is not specified for the model")
        model = {
            "name": self.name,
            "signature": ModelSignature_to_signature_dict(self.signature),
            "metadata": self.metadata
        }
        resp = cluster.request("POST", "/api/v2/externalmodel", json=model)
        handle_request_error(
            resp, f"Failed to create an external model. {resp.status_code} {resp.text}")
        json_dict = resp.json()
        json_dict['isExternal'] = True
        return ModelVersion._from_json(cluster, json_dict)


class ModelVersionStatus(str, Enum):
    """
    Model building statuses.
    """
    Assembling = "Assembling"
    Released = "Released"
    Failed = "Failed"


class ModelVersion:
    """
    A Model, registered within the cluster. ModelVersion represents one of the Models' versions.
    (https://hydrosphere.io/serving-docs/latest/overview/concepts.html#models)

    :Example:

    List all modelversions, registered on the cluster.

    >>> from hydrosdk.cluster import Cluster
    >>> cluster = Cluster("http-cluster-endpoint")
    >>> for model_version in ModelVersion.list(cluster):
            print(model_version)

    Upload training data for a model_version.

    >>> from hydrosdk.cluster import Cluster
    >>> cluster = Cluster("http-cluster-endpoint")
    >>> model_version = ModelVersion.find(cluster, "my-model", 1)
    >>> model_version.lock_till_released()
    >>> model_version.training_data = "s3://my-bucket/path/to/training-data.csv"
    >>> data_upload_response = model_version.upload_traininf_data()

    Assign custom monitoring metrics for a model_version.

    >>> from hydrosdk.cluster import Cluster
    >>> from hydrosdk.monitoring import ThresholdCmpOp
    >>> cluster = Cluster("http-cluster-endpoint")
    >>> model_version = ModelVersion.find(cluster, "my-model", 1)
    >>> modelversion_metric = ModelVersion.find(cluster, "my-model-metric", 1)
    >>> modelversion_metric = modmodelversion_metric.as_metric(1.4, ThresholdCmpOp.LESS_EQ)
    >>> model_version.assign_metrics([modelversion_metric])

    Check logs from a ModelVersion.

    >>> from hydrosdk.cluster import Cluster
    >>> cluster = Cluster("http-cluster-endpoint")
    >>> model_version = ModelVersion.find(cluster, "my-model", 1)
    >>> for event in model_version.logs():
            print(event.data)
    """
    _BASE_URL = "/api/v2/model"

    @staticmethod
    def list(cluster: Cluster) -> List['ModelVersion']:
        """
        List all model versions on the cluster.

        :param cluster: active cluster
        :return: list of ModelVersions 
        """

        resp = cluster.request("GET", f"{ModelVersion._BASE_URL}/version")
        handle_request_error(
            resp, f"Failed to list model versions. {resp.status_code} {resp.text}")
        return [ModelVersion._from_json(cluster, modelversion_json)
                for modelversion_json in resp.json()]
    
    @staticmethod
    def find(cluster: Cluster, name: str, version: int) -> 'ModelVersion':
        """
        Find a ModelVersion on the cluster by model name and a version.

        :param cluster: active cluster
        :param name: name of the model
        :param version: version of the model
        :return: ModelVersion object
        """
        resp = cluster.request("GET", f"{ModelVersion._BASE_URL}/version/{name}/{version}")
        handle_request_error(
            resp, f"Failed to find model_version for name={name}, version={version}. {resp.status_code} {resp.text}")
        return ModelVersion._from_json(cluster, resp.json())

    @staticmethod
    def find_by_id(cluster: Cluster, id: int) -> 'ModelVersion':
        """
        Find a ModelVersion on the cluster by id.

        :param cluster: active cluster
        :param id: model version id
        :return: ModelVersion object
        """
        resp = cluster.request("GET", f"{ModelVersion._BASE_URL}/version")
        handle_request_error(
            resp, f"Failed to retrieve model_version by id={id}. {resp.status_code} {resp.text}")
        for modelversion_json in resp.json():
            if modelversion_json['id'] == id:
                return ModelVersion._from_json(cluster, modelversion_json)
        raise BadRequestException(f"Failed to retrieve model_version by id={id}.")
    
    @staticmethod
    def find_by_model_name(cluster: Cluster, model_name: str) -> list:
        """
        Find all model versions on the cluster filtered by model_name, sorted in ascending 
        order by version.

        :param cluster: active cluster
        :param model_name: a model name
        :return: list of ModelVersions with `model_name`
        """
        all_models = ModelVersion.list(cluster=cluster)
        modelversions_by_name = [model for model in all_models if model.name == model_name]
        sorted_by_version = sorted(modelversions_by_name, key=lambda model: model.version)
        return sorted_by_version

    @staticmethod
    def _from_json(cluster: Cluster, modelversion_json: dict) -> 'ModelVersion':
        """
        Internal method used for deserealization of a ModelVersion from a json object.

        :param cluster: active cluster
        :param modelversion_json: json response from the cluster
        :return: ModelVersion object
        """
        id = modelversion_json["id"]
        name = modelversion_json["model"]["name"]
        model_id = modelversion_json["model"]["id"]
        version = modelversion_json["modelVersion"]
        signature = signature_dict_to_ModelSignature(modelversion_json["modelSignature"])
        monitoring_configuration = MonitoringConfiguration(batch_size=modelversion_json["monitoringConfiguration"]["batchSize"])

        # external model deserialization handling
        is_external = modelversion_json.get('isExternal', False)
        if is_external:
            model_runtime = None
        else:
            model_runtime = DockerImage(**modelversion_json["runtime"])
        
        model_image = modelversion_json.get("image")
        if model_image is not None:
            model_image = DockerImage(**model_image)

        status = modelversion_json.get('status')
        if status:
            status = ModelVersionStatus[status]
        metadata = modelversion_json['metadata']

        return ModelVersion(
            id=id,
            model_id=model_id,
            is_external=is_external,
            name=name,
            version=version,
            signature=signature,
            runtime=model_runtime,
            image=model_image,
            cluster=cluster,
            status=status,
            metadata=metadata,
            monitoring_configuration=monitoring_configuration,
            created=modelversion_json.get("created"),
            finished=modelversion_json.get("finished"),
            applications=modelversion_json.get("applications"),
        )

    def lock_till_released(self, timeout: int = 120):
        """
        Lock till the model completes assembling.
        
        :param timeout: approximate value when to raise TimeoutException
        :raises ModelVersion.ReleaseFailed: if model failed to be released
        :raises TimeoutException: if timed out waiting for the model to be released
        """
        events_stream = self.cluster.request("GET", "/api/v2/events", stream=True)
        events_client = sseclient.SSEClient(events_stream)

        self.status = self.find_by_id(self.cluster, self.id).status
        if self.status is ModelVersionStatus.Released:
            return None
        if self.status is ModelVersionStatus.Failed:
            raise ModelVersion.ReleaseFailed("Model version build failed.")
        try:
            deadline_at = datetime.datetime.now().timestamp() + timeout
            for event in events_client.events():
                if datetime.datetime.now().timestamp() > deadline_at:
                    # last chance
                    self.status = self.find_by_id(self.cluster, self.id).status
                    if self.status is ModelVersionStatus.Released:
                        return None
                    raise TimeoutException("Timed out waiting for the model version to build.")
                if event.event == "ModelUpdate":
                    data = json.loads(event.data)
                    if data.get("id") == self.id:
                        self.status = ModelVersionStatus[data.get('status')]
                        if self.status is ModelVersionStatus.Released:
                            return None
                        raise ModelVersion.ReleaseFailed("Model version build failed.")
        finally:
            events_client.close()
    
    def build_logs(self) -> Iterator[Event]:
        """
        Sends request, saves and returns a build logs iterator.

        :return: Iterator over sseclient.Event
        """
        resp = self.cluster.request("GET", f"{self._BASE_URL}/version/{self.id}/logs", stream=True)
        return sseclient.SSEClient(resp).events()

    def upload_training_data(self) -> 'DataUploadResponse':
        """
        Uploads training data for a given model_version.

        :raises: ValueError if training data is not specified
        """
        if self.training_data is not None:
            return _upload_training_data(self.cluster, self.id, self.training_data)
        raise ValueError('Training data is not specified')

    def assign_metrics(self, metrics: List[MetricModel], wait: bool = True):
        """
        Adds metrics to the model.

        :param metrics: list of metrics
        :return: self
        """
        if wait:
            self.lock_till_released()
        
        for metric in metrics:
            modelversion = metric.modelversion
            if wait:
                modelversion.lock_till_released()

            config = MetricSpecConfig(
                modelversion_id=modelversion.id,
                threshold=metric.threshold,
                threshold_op=metric.comparator
            )
            MetricSpec.create(
                cluster=self.cluster,
                name=modelversion.name,
                modelversion_id=self.id,
                config=config
            )

    def to_proto(self) -> ModelVersionProto:
        """
        Convert ModelVersion object into a ModelVersion proto message.

        :return: ModelVersion proto message
        """
        return ModelVersionProto(
            _id=self.id,
            name=self.name,
            version=self.version,
            signature=self.signature,
            runtime=self.runtime,
            image=DockerImageProto(name=self.image.full, tag=self.image.tag),
            image_sha=self.image.sha256
        )
    
    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
            "modelVersion": self.version,
            "modelSignature": ModelSignature_to_signature_dict(self.signature),
            "runtime": self.runtime._asdict(),
            "image": self.image._asdict(),
            "status": self.status,
            "metadata": self.metadata,
            "monitoringConfiguration": self.monitoring_configuration.to_dict(),
            "isExternal": self.is_external,
            "created": self.created,
            "finished": self.finished,
            "applications": self.applications,
        }
    
    def as_metric(self, threshold: int, comparator: ThresholdCmpOp) -> MetricModel:
        """
        Converts model to MetricModel.

        :param threshold:
        :param comparator:
        :return: MetricModel
        """
        return MetricModel(modelversion=self, threshold=threshold, comparator=comparator)

    def __init__(self, cluster: Cluster, id: int, model_id: int, name: str, version: int, 
                 signature: ModelSignature, status: Optional[ModelVersionStatus], image: Optional[DockerImage], 
                 runtime: Optional[DockerImage], is_external: bool, 
                 metadata: Optional[Dict[str, str]] = None, install_command: Optional[str] = None, 
                 training_data: Optional[str] = None, monitoring_configuration: Optional[MonitoringConfiguration] = None, 
                 created: Optional[str] = None, finished: Optional[str] = None, applications: Optional[List[str]] = None):
        """
        :param cluster: active cluster
        :param id: id of the model_version assigned by the cluster
        :param model_id: id of the model assigned by the cluster
        :param name: a name of the model this ModelVersion belongs to
        :param version: a version of the model this ModelVersion belongs to
        :param signature: a ModelSignature which specifies the name of the inference function, 
                          as well as its types and shapes for both inputs and outputs
        :param status: a status of this ModelVersion, one of {Assembling, Released, Failed}
        :param image: DockerImage of a packed ModelVersion stored inside the cluster
        :param runtime: DockerImage of a runtime which was used to build an `image`
        :param is_external: indicates whether model is running outside the cluster
        :param metadata: metadata used for describing a ModelVersion
        :param install_command: a command which was run within a runtime to prepare a model_version
                                environment
        :param monitoring_configuration: specified monitoring configuration for the model
        :param created: specifies, when the model version was created
        :param finished: specified, when the model has finished building
        :param applications: specifies, which applications are using this model version
        """
        self.id = id
        self.model_id = model_id
        self.name = name
        self.runtime = runtime
        self.is_external = is_external
        self.signature = signature
        self.cluster = cluster
        self.version = version
        self.image = image
        self.status = status
        self.metadata = metadata
        self.install_command = install_command
        self.training_data = training_data
        self.monitoring_configuration = monitoring_configuration
        self.created = created
        self.finished = finished
        self.applications = applications or []
        if self.cluster.channel:
            self.analyze_stub = MonitoringServiceStub(self.cluster.channel) 

    def __repr__(self):
        return f"{self.name}:{self.version}"

    class ReleaseFailed(HydrosphereException):
        pass

    def analyze(self, *, request_id: str, request: PredictRequest, response: Optional[PredictResponse]=None, error: Optional[str]=None):
        """
        Sends inference data to monitoring service. Used if model is deployed in external inference service.
        """
        return _analyze(self, self.analyze_stub, request_id, request, response, error)


class DataProfileStatus(Enum):
    Success = "Success"
    Failure = "Failure"
    Processing = "Processing"
    NotRegistered = "NotRegistered"
    Unknown = "Unknown"


# TODO: This should be rewritten to SSE, similar to lock_* methods from other entities, 
# like ModelVersion.lock_till_released or Application.lock_while_starting

class DataUploadResponse:
    """
    Handle processing status of the training data upload. 

    :Example:

    Wait till data processing gets finished.

    >>> from hydrosdk.cluster import Cluster
    >>> cluster = Cluster("http-cluster-endpoint")
    >>> model_version = ModelVersion.find(cluster, "my-model", 1)
    >>> model_version.training_data = "s3://bucket/path/to/training-data.csv"
    >>> data_upload_response = model_version.upload_training_data()
    >>> try: 
            data_upload_response.wait(retry=3, sleep=30) 
        except TimeoutException as e:
            print(e)
        except DataUploadResponse.NotRegistered:
            id = data_upload_response.modelversion_id
            print(f"Unable to find training data for modelversion_id={id}")
        except DataUploadResponse.Failed:
            print(f"Failed to process training data")
    """
    def __init__(self, cluster: Cluster, modelversion_id: int) -> 'DataUploadResponse':
        self.cluster = cluster
        self.modelversion_id = modelversion_id

    @property
    def url(self) -> str:
        if not hasattr(self, '__url'):
            self.__url = f'/monitoring/profiles/batch/{self.modelversion_id}/status'
        return self.__url
    
    def get_status(self) -> DataProfileStatus:
        resp = self.cluster.request('GET', self.url)
        handle_request_error(
            resp, f"Failed to get status for modelversion_id={self.modelversion_id}. {resp.status_code} {resp.text}")
        return DataProfileStatus[resp.json().get('kind', "Unknown")]

    def wait(self, retry=12, sleep=30):
        """
        Wait till data processing gets finished. 
        
        :param retry: Number of retries before giving up waiting.
        :param sleep: Sleep interval between retries.
        """
        def tick(retry: int, sleep: int) -> Tuple[bool, int]:
            if retry == 0:
                return False, retry
            retry -= 1
            time.sleep(sleep)
            return True, retry
            
        while True:
            status = self.get_status()
            if status is DataProfileStatus.Success:
                break
            elif status is DataProfileStatus.Processing:
                can_be_polled, retry = tick(retry, sleep)
                if can_be_polled:
                    continue
                raise TimeoutException("Time out waiting for data processing to complete")
            elif status is DataProfileStatus.NotRegistered:
                raise DataUploadResponse.NotRegistered(f"Data profiling for modelversion_id={self.modelversion_id} has not been registered.")
            elif status is DataProfileStatus.Unknown:
                raise DataUploadResponse.Unknown(f"Data profiling for modelversion_id={self.modelversion_id} is unknown.")
            raise DataUploadResponse.Failed(f"Data profiling for modelversion_id={self.modelversion_id} failed.")

    class NotRegistered(HydrosphereException):
        pass

    class Failed(HydrosphereException):
        pass

    class Unknown(HydrosphereException):
        pass



from enum import Enum
from typing import List, Optional, Dict
import json
import datetime

import sseclient
from google.protobuf.json_format import MessageToDict

from hydro_serving_grpc.serving.contract.signature_pb2 import ModelSignature

from hydrosdk.cluster import Cluster
from hydrosdk.signature import signature_dict_to_ModelSignature
from hydrosdk.data.types import PredictorDT
from hydrosdk.deployment_configuration import DeploymentConfiguration
from hydrosdk.modelversion import ModelVersion
from hydrosdk.predictor import PredictServiceClient, MonitorableApplicationPredictionService
from hydrosdk.utils import handle_request_error
from hydrosdk.exceptions import TimeoutException
from pydantic import BaseModel

class StreamingParams(BaseModel):
    sourceTopic: str
    destinationTopic: str

class ModelVariant(BaseModel):
    modelVersionId: int
    weight: int
    deploymentConfigurationName: Optional[str]
    servableName: Optional[str]


class ApplicationStatus(Enum):
    FAILED = 0
    ASSEMBLING = 1
    READY = 2


class Application:
    """
    Applications are used to combine your ModelVersions into a linear graph and deploy 
    it into production, exposing HTTP and gRPC interfaces for consumers.

    Use ApplicationBuilder class to create a new Application in your Hydrosphere cluster
    or an Application.find method to get the existing Application.

    :Example:

    List all applications created on the cluster.

    >>> from hydrosdk.cluster import Cluster 
    >>> cluster = Cluster("http-cluster-endpoint")
    >>> apps = Application.list(cluster)
    >>> for app in apps: 
    >>>     print(app)

    Find an application by name and perform a prediction from it.

    >>> from hydrosdk.cluster import Cluster
    >>> cluster = Cluster("http-cluster-endpoint", "grpc-cluster-endpoint")  # important to use a gRPC endpoint 
    >>> app = Application.find(cluster, "my-application")
    >>> pred = app.predictor()
    >>> resp = pred.predict({"my-input": 1})
    """
    _BASE_URL = "/api/v2/application"

    @staticmethod
    def list(cluster: Cluster) -> List['Application']:
        """
        List all available applications from the cluster.

        :param cluster: active cluster
        :return: deserialized list of application objects
        """
        resp = cluster.request("GET", Application._BASE_URL)
        handle_request_error(
            resp, f"Failed to list all applications. {resp.status_code} {resp.text}")
        applications = [Application._from_json(cluster, app_json)
                        for app_json in resp.json()]
        return applications

    @staticmethod
    def find(cluster: Cluster, application_name: str) -> 'Application':
        """
        Search for an application by name. 

        :param cluster: active cluster
        :param application_name: application name
        :return: deserialized application object
        """
        resp = cluster.request("GET", f"{Application._BASE_URL}/{application_name}")
        handle_request_error(
            resp, f"Failed to find an application by name={application_name}. {resp.status_code} {resp.text}")
        return Application._from_json(cluster, resp.json())

    @staticmethod
    def delete(cluster: Cluster, application_name: str) -> dict:
        """
        Delete an application by name.

        :param cluster: active cluster
        :param application_name: application name
        :return: response from the cluster
        """
        resp = cluster.request("DELETE", f"{Application._BASE_URL}/{application_name}")
        handle_request_error(
            resp, f"Failed to delete application for name={application_name}. {resp.status_code} {resp.text}")
        return resp.json()

    @staticmethod
    def _from_json(cluster: Cluster, application_json: dict) -> 'Application':
        """
        Deserialize json into application object. 

        :param cluster: active cluster
        :param application_json: input json with application object fields
        :return: application object
        """
        id_ = application_json.get("id")
        name = application_json.get("name")
        execution_graph = ExecutionGraph._from_json(cluster, application_json.get("executionGraph"))
        kafka_streaming = [StreamingParams(sourceTopic=kafka_param["in-topic"], destinationTopic=kafka_param["out-topic"])
                               for kafka_param in application_json.get("kafkaStreaming")]
        metadata = application_json.get("metadata")
        message = application_json.get("message")
        signature = signature_dict_to_ModelSignature(data=application_json.get("signature"))
        status = ApplicationStatus[application_json.get("status").upper()]

        app = Application(cluster=cluster,
                          id=id_,
                          name=name,
                          execution_graph=execution_graph,
                          status=status,
                          signature=signature,
                          kafka_streaming=kafka_streaming,
                          metadata=metadata,
                          message=message)
        return app

    def lock_while_starting(self, timeout: int = 120) -> 'Application':
        """ Wait for an application to become ready. """
        events_stream = self.cluster.request("GET", "/api/v2/events", stream=True)
        events_client = sseclient.SSEClient(events_stream)

        self.status = self.find(self.cluster, self.name).status
        if self.status is ApplicationStatus.READY: 
            return self
        if self.status is ApplicationStatus.FAILED:
            raise ValueError(f'Application initialization failed {self.message}')
        try:
            deadline_at = datetime.datetime.now().timestamp() + timeout
            for event in events_client.events():
                if datetime.datetime.now().timestamp() > deadline_at:
                    raise TimeoutException('Time out waiting for an application to become available')
                if event.event == "ApplicationUpdate":
                    data = json.loads(event.data)
                    print(data)
                    if data.get("name") == self.name:
                        self.status = ApplicationStatus[data.get("status").upper()]
                        if self.status is ApplicationStatus.READY:
                            return self
                        elif self.status is ApplicationStatus.FAILED:
                            raise ValueError('Application initialization failed')
        finally:
            events_client.close()

    def predictor(self, return_type=PredictorDT.DICT_NP_ARRAY) -> PredictServiceClient:
        """
        Return a predictor object which is used to transform your data
        into a proto message, pass it via gRPC to the cluster and decode
        the cluster output from proto to a dict with Python dtypes.

        :param return_type: Specifies into which format should predictor serialize model output.
                            Numpy dtypes, Python dtypes or pd.DataFrame are supported.
        :return: PredictServiceClient with .predict() method which accepts your data
        """
        impl = MonitorableApplicationPredictionService(channel=self.cluster.channel, target=self.name)
        return PredictServiceClient(impl=impl, signature=self.signature, return_type=return_type)

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "signature": MessageToDict(self.signature),
            "execution_graph": self.execution_graph.to_dict(),
            "metadata": self.metadata,
            "message": self.message,
            "status": self.status.name,
        }
    
    def __init__(self, cluster: Cluster, id: int, name: str, execution_graph: 'ExecutionGraph',
                 status: ApplicationStatus, signature: ModelSignature,
                 kafka_streaming: List[StreamingParams],
                 metadata: Optional[Dict[str, str]] = None,
                 message: Optional[str] = None) -> 'Application':
        """
        :param cluster: active cluster
        :param id: unique application ID
        :param name: application Name
        :param signature: signature, specifying input and output fields names, dtypes and shapes
        :param execution_graph: linear graph which specifies ExecutionStages which sequentially transform input
        :param status: Application Status
        :param kafka_streaming: list of Kafka parameters with input and output Kafka topics specified
        :param metadata: metadata with string keys and string values.
        :param message: possible error message from the cluster
        """
        self.id = id
        self.name = name
        self.execution_graph = execution_graph
        self.kafka_streaming = kafka_streaming
        self.metadata = metadata
        self.status = status
        self.signature = signature
        self.cluster = cluster
        self.message = message

    def __str__(self):
        return f"Application {self.id} {self.name}"


class ApplicationBuilder:
    """
    ApplicationBuilder is used to create new Applications in your cluster.

    :Example:
    
    Create an application from existing modelversions.

    >>> from hydrosdk import Cluster, ModelVersion
    >>> cluster = Cluster('http-cluster-endpoint')
    >>> mv1 = ModelVersion.find(cluster, "my-model", 1)
    >>> mv2 = ModelVersion.find(cluster, "my-model", 2)
    >>> stage = ExecutionStageBuilder() \
                    .with_model_variant(mv1, 50) \
                    .with_model_variant(mv2, 50) \ 
                    .build()
    >>> app = ApplicationBuilder("my-application-ab-test") \
                .with_stage(stage) \
                .build(cluster)
    """

    def __init__(self, name: str) -> 'ApplicationBuilder':
        """
        :param cluster: Hydrosphere cluster where you want to create an Application
        :param name: Future Application name
        """
        self.name = name
        self.stages = []
        self.metadata = {}
        self.streaming_parameters = []

    def with_stage(self, stage: 'ExecutionStage') -> 'ApplicationBuilder':
        """
        Add an ExecutionStage to your Application. See ExecutionStage for more information.

        :param stage:
        :return:
        """
        self.stages.append(stage)
        return self

    def with_metadata(self, key: str, value: str) -> 'ApplicationBuilder':
        """
        Add a metadata value to your future Application.

        :param key: string key under which `value` will be stored
        :param value: string value
        :return:
        """
        self.metadata[key] = value
        return self

    def with_metadatas(self, metadata: Dict[str, str]) -> 'ApplicationBuilder':
        """
        Add a metadata to your future Application.

        :param metadata: a dict containing metadata for your application
        :return:
        """
        self.metadata.update(metadata)
        return self

    def with_kafka_params(self, source_topic: str, dest_topic: str) -> 'ApplicationBuilder':
        """
        Add a kafka parameters to your Application.

        :param source_topic: source Kafka topic
        :param dest_topic: destination Kafka topic
        :return:
        """
        params = StreamingParams(sourceTopic=source_topic, destinationTopic=dest_topic)
        self.streaming_parameters.append(params)
        return self

    def build(self, cluster: Cluster) -> Application:
        """
        Create an Application in your Hydrosphere cluster.

        :return: Application object
        """
        if not self.stages:
            raise ValueError("No execution stages were provided")

        execution_graph = ExecutionGraph(stages=self.stages)
        application_json = {"name": self.name,
                            "kafkaStreaming": [sp.to_dict() for sp in self.streaming_parameters],
                            "executionGraph": execution_graph.to_dict(),
                            "metadata": self.metadata}

        resp = cluster.request("POST", Application._BASE_URL, json=application_json)
        handle_request_error(
            resp, f"Failed to create an application {self.name}. {resp.status_code} {resp.text}")
        return Application._from_json(cluster, resp.json())


class ExecutionGraph:
    def __init__(self, stages: List['ExecutionStage']) -> 'ExecutionGraph':
        """
        ExecutionGraph is a representation of a linear graph which is used
        by Hydrosphere to create pipelines of ModelVersions. This linear graph
        consists of ExecutionStages following each other. See ExecutionStage to learn more.

        :param stages: list of ExecutionStages used to sequentially transform input.
        """
        self.stages = stages

    def to_dict(self) -> Dict[str, List]:
        return {"stages": [s.to_dict() for s in self.stages]}

    @staticmethod
    def _from_json(cluster: Cluster, ex_graph_dict: Dict) -> 'ExecutionGraph':
        return ExecutionGraph([ExecutionStage._from_json(cluster, stage)
                               for stage in ex_graph_dict['stages']])


class ExecutionStage:
    def __init__(self, model_variants: List[ModelVariant], signature: Optional[ModelSignature]) -> 'ExecutionStage':
        """
        ExecutionStage is a single stage in a linear graph of ExecutionGraph. Each stage
        may contain from 1 to many different ModelVersions with the same signature. Every input
        requested routed to this stage will be automatically shadowed to all ModelVersions inside
        of it. Stage output response will be selected according to the relative weights associated 
        with each version.

        :param model_variants: list of ModelVersions with corresponding weights
        :param signature: signature specifying input and output field names, data types and shapes.
        """
        self.signature = signature
        self.model_variants = model_variants

    def to_dict(self) -> Dict[str, List[Dict[str, int]]]:
        model_variants = []
        for model_variant in self.model_variants:
            dict_repr = {
                'modelVersionId': model_variant.modelVersionId,
                'weight': model_variant.weight,
                'deploymentConfigName': model_variant.deploymentConfigurationName,
            }
            model_variants.append(dict_repr)
        return {"modelVariants": model_variants}

    @staticmethod
    def _from_json(cluster: Cluster, execution_stage_dict: Dict) -> 'ExecutionStage':
        execution_stage_signature = signature_dict_to_ModelSignature(execution_stage_dict['signature'])
        model_variants = [ModelVariant.parse_obj(mv) for mv in execution_stage_dict['modelVariants']]
        return ExecutionStage(model_variants=model_variants, signature=execution_stage_signature)


class ExecutionStageBuilder:
    def __init__(self) -> 'ExecutionStageBuilder':
        """
        Builder class to help building ExecutionStage.
        """
        self.model_variants = []

    def __validate(self):
        """
        Validate the stage for correctness.
        """
        if len(self.model_variants) == 0:
            raise ValueError("At least one model variant should be specified.")

        if sum(variant.weight for variant in self.model_variants) != 100:
            raise ValueError("All model variants' weights inside the same stage must sum up to 100")

    def with_model_variant(self, model_version: ModelVersion,
                           weight: int,
                           deployment_configuration: Optional[DeploymentConfiguration] = None) -> 'ExecutionStageBuilder':
        """
        Add a ModelVersion with a weight to an ExecutionStage.
        
        :param model_version: ModelVersion to which input requests will be shadowed
        :param weight: Weight which affects the chance of choosing a model output as an
         output of an ExecutionStage
        :param deployment_configuration: K8s Deployment Configuration of this Model Variant
        :return:
        """
        dc_name = None
        if deployment_configuration is not None:
            dc_name = deployment_configuration.name
        mv = ModelVariant(
            modelVersionId=model_version.id,
            weight=weight,
            deploymentConfigurationName=dc_name,
            servableName=None
        )
        self.model_variants.append(mv)
        return self

    def build(self) -> 'ExecutionStage':
        """
        Verify that all ModelVersions inside an ExecutionStage have the same signature
        and finally creates an ExecutionStage.

        :return:
        """
        self.__validate()
        return ExecutionStage(model_variants=self.model_variants, signature=None)



from typing import Dict, List, Iterable, Union

import numpy as np
import pandas as pd
from hydro_serving_grpc.serving.contract.tensor_pb2 import Tensor, TensorShape
from hydro_serving_grpc.serving.contract.types_pb2 import DT_STRING, DT_HALF, DT_COMPLEX64, DT_COMPLEX128, DataType
from hydro_serving_grpc.serving.contract.signature_pb2 import ModelSignature
from pandas.core.common import flatten

from hydrosdk.data.types import np_to_proto_dtype, DTYPE_TO_FIELDNAME, find_in_list_by_name, proto_to_np_dtype


def tensor_proto_to_py(t: Tensor):
    """
    Converts tensor proto into a corresponding python object - list or scalar
    :param t:
    :return:
    """
    dims = list(t.tensor_shape.dims)
    value = getattr(t, DTYPE_TO_FIELDNAME[t.dtype])

    # If no dims specified in TensorShape, then it is scalar
    if dims:
        value = np.reshape(value, dims).tolist()
        return value
    else:
        return value[0]


def list_to_tensor_proto(data: List, proto_dtype: DataType, proto_shape: TensorShape) -> Tensor:
    """
    Converts data in a form of a Python List into a Tensor object
    :param data: List with data
    :param proto_dtype: DataType of a future Tensor
    :param proto_shape: TensorShape of a future Tensor
    :return: Same data but in a Tensor object
    """
    # We can pack only flattened lists into Tensor, so we need to flatten the list
    flattened_list = flatten(data)
    tensor_proto_parameters = {
        DTYPE_TO_FIELDNAME[proto_dtype]: flattened_list,
        "dtype": proto_dtype,
        "tensor_shape": proto_shape
    }
    return Tensor(**tensor_proto_parameters)


def tensor_proto_to_np(t: Tensor):
    """
    Creates either np.array or scalar with Numpy dtype based on
     data type, shape and values from Tensor object
    :param t:
    :return:
    """
    array_shape = list(t.tensor_shape.dims)
    np_dtype = proto_to_np_dtype(t.dtype)
    proto_values = getattr(t, DTYPE_TO_FIELDNAME[t.dtype])

    if t.dtype == DT_HALF:
        x = np.fromiter(proto_values, dtype=np.uint16).view(np.float16)
    elif t.dtype == DT_STRING:
        x = np.array([s.decode("utf-8") for s in proto_values])
    elif t.dtype == DT_COMPLEX64 or t.dtype == DT_COMPLEX128:
        it = iter(proto_values)
        x = np.array([complex(x[0], x[1]) for x in zip(it, it)], dtype=np_dtype)
    else:
        x = np.array(proto_values, dtype=np_dtype)

    # If no dims specified in TensorShape, then it is scalar
    if array_shape:
        return x.reshape(*array_shape)
    else:
        return x.flatten()[0]


def np_to_tensor_proto(x) -> Tensor:
    """
    Creates Tensor object from Numpy ndarray or scalar with inferred TensorProtoShape and DataType
    :param x: Union[np.array, np.ScalarType]
    :return:
    """
    if isinstance(x, np.ScalarType):
        return scalar_to_tensor_proto(x)
    elif isinstance(x, np.ndarray):
        return nparray_to_tensor_proto(x)
    else:
        raise TypeError(f"Unsupported object {x}")


def nparray_to_tensor_proto(x: np.array) -> Tensor:
    """
    Creates Tensor object from Numpy ndarray
    with TensorProtoShape and DataType inferred from the latter
    :param x: Data in form ofa numpy ndarray
    :return: Same data packed into a Tensor object
    """

    if x.dtype.isbuiltin != 1 and x.dtype.type != np.str_:
        raise ValueError(f"{x.dtype} is not supported."
                         f" Dtypes not compiled into numpy are not supported, except for np.str.")

    proto_dtype = np_to_proto_dtype(x.dtype.type)
    proto_shape = tensor_shape_proto_from_list(x.shape)

    if proto_dtype == DT_HALF:
        proto_values = x.view(np.uint16).flatten()
    elif proto_dtype == DT_STRING:
        proto_values = [s.encode("utf-8") for s in x.flatten()]
    elif proto_dtype == DT_COMPLEX64 or proto_dtype == DT_COMPLEX128:
        proto_values = [v.item() for c_number in x.flatten() for v in [c_number.real, c_number.imag]]
    else:
        proto_values = x.flatten()

    kwargs = {
        DTYPE_TO_FIELDNAME[proto_dtype]: proto_values,
        "dtype": proto_dtype,
        "tensor_shape": proto_shape
    }

    return Tensor(**kwargs)


def scalar_to_tensor_proto(x: np.ScalarType) -> Tensor:
    """
    Creates Tensor object from a scalar with a Numpy dtype
      with TensorProtoShape and DataType inferred from the latter
    :param x: Scalar value with a Numpy dtype
    :return: Same value but packed into a Tensor object
    """
    proto_dtype = np_to_proto_dtype(type(x))

    if proto_dtype == DT_HALF:
        proto_values = [np.array(x, dtype=np.float16).view(np.uint16)]
    elif proto_dtype == DT_STRING:
        proto_values = [x.encode("utf-8")]
    elif proto_dtype == DT_COMPLEX64 or proto_dtype == DT_COMPLEX128:
        proto_values = [x.real, x.imag]
    else:
        proto_values = [x]

    kwargs = {
        DTYPE_TO_FIELDNAME[proto_dtype]: proto_values,
        "dtype": proto_dtype,
        "tensor_shape": TensorShape()
    }
    return Tensor(**kwargs)


def tensor_shape_proto_from_list(shape: Iterable[int]) -> TensorShape:
    """
    Helper function to transform shape in the form of a tuple (Numpy shape representation) into a TensorProtoShape
    :param shape: Shape in a tuple form
    :return: same shape but in a TensorShape object
    """
    return TensorShape(dims=shape)


def isinstance_namedtuple(obj) -> bool:
    """
    Based on https://stackoverflow.com/a/49325922/7127824 and https://github.com/Hydrospheredata/hydro-serving-sdk/pull/51
    The main use is to check for namedtuples returned by pandas.DataFrame.itertuples()

    :param obj: any object
    :return: bool if object is an instance of namedtuple
    """

    return (isinstance(obj, tuple) and
            hasattr(obj, '_asdict') and
            hasattr(obj, '_fields'))


def convert_inputs_to_tensor_proto(inputs: Union[pd.DataFrame, dict, pd.Series], signature: ModelSignature) -> Dict[str, Tensor]:
    """
    Generate Dict[str, Tensor] from pd.DataFrame or Dict[str, Union[np.array, np.ScalarType]]

    Converts inputs into a representation of data where each field
     of a signature is represented by a valid Tensor object.
    :param inputs: Dict, where keys are names of signature fields and
     values are data in either Numpy or Python form, or alternatively,
     pd.DataFrame, where columns are names of fields and column values are data.
    :param signature: ModelVersion signature with names, shapes and dtypes
     of fields into which `inputs` are converted into
    :return: Dictionary with TensorProtos to be used in forming a PredictRequest
    """
    tensors = {}

    # if we get a namedtuple, we pass further dict
    if isinstance_namedtuple(inputs):
        inputs = inputs._asdict()

    if isinstance(inputs, dict):
        for key, value in inputs.items():
            if isinstance(value, list):  # x: [1,2,3,4]
                signature_field = find_in_list_by_name(some_list=signature.inputs, name=key)
                tensors[key] = list_to_tensor_proto(value, signature_field.dtype, signature_field.shape)
            elif isinstance(value, np.ScalarType):
                # This works for all scalars, including python int, float, etc.
                tensors[key] = scalar_to_tensor_proto(value)
            elif isinstance(value, np.ndarray):
                tensors[key] = nparray_to_tensor_proto(value)
            else:
                raise TypeError("Unsupported objects in dict values {}".format(type(value)))
    elif isinstance(inputs, pd.DataFrame):
        for key, value in dict(inputs).items():
            tensors[key] = nparray_to_tensor_proto(value.ravel())
    elif isinstance(inputs, pd.Series):
        for key, value in inputs.items():
            tensors[key] = scalar_to_tensor_proto(value)
    else:
        raise ValueError(f"Conversion failed. Expected [pandas.DataFrame, dict[str, numpy.ndarray],\
                           dict[str, list], dict[str, np.ScalarType]], got {type(inputs)}")

    return tensors
